"""
Module for managing an climate mode remote values.

DPT .
"""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, TypeVar

from xknx.dpt import (
    DPTArray,
    DPTBinary,
    DPTHVACContrMode,
    DPTHVACMode,
    DPTHVACStatus,
)
from xknx.dpt.dpt_1 import HeatCool
from xknx.dpt.dpt_20 import HVACControllerMode, HVACOperationMode, HVACStatus
from xknx.exceptions import ConversionError, CouldNotParseTelegram

from .remote_value import (
    GroupAddressesType,
    RemoteValue,
    RVCallbackType,
)

if TYPE_CHECKING:
    from xknx.xknx import XKNX


HVACModeT = TypeVar(
    "HVACModeT",
    HVACControllerMode,
    HVACOperationMode,
    HVACStatus,
    HVACOperationMode | None,
)


class RemoteValueClimateModeBase(RemoteValue[HVACModeT]):
    """Base class for binary climate mode remote values."""

    @abstractmethod
    def supported_operation_modes(self) -> list[HVACOperationMode]:
        """Return a list of all supported operation modes."""

    @abstractmethod
    def set_operation_mode(self, mode: HVACOperationMode) -> None:
        """Set operation mode. Return if not supported."""

    @abstractmethod
    def supported_controller_modes(self) -> list[HVACControllerMode]:
        """Return a list of all supported controller modes."""

    @abstractmethod
    def set_controller_mode(self, mode: HVACControllerMode) -> None:
        """Set controller mode. Return if not supported."""


class RemoteValueOperationMode(RemoteValueClimateModeBase[HVACOperationMode]):
    """Abstraction for remote value of KNX climate operation modes."""

    dpt_class = DPTHVACMode

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "Climate mode",
        after_update_cb: RVCallbackType[HVACOperationMode] | None = None,
    ) -> None:
        """Initialize remote value of KNX climate mode."""
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    def supported_operation_modes(self) -> list[HVACOperationMode]:
        """Return a list of all supported operation modes."""
        return self.dpt_class.get_valid_values()

    def set_operation_mode(self, mode: HVACOperationMode) -> None:
        """Set operation mode. Return if not supported."""
        if mode not in self.supported_operation_modes():
            return
        super().set(mode)

    def supported_controller_modes(self) -> list[HVACControllerMode]:
        """Return a list of all supported controller modes."""
        return []

    def set_controller_mode(self, mode: HVACControllerMode) -> None:
        """Set controller mode. Return if not supported."""
        return


class RemoteValueControllerMode(RemoteValueClimateModeBase[HVACControllerMode]):
    """Abstraction for remote value of KNX climate controller modes."""

    dpt_class = DPTHVACContrMode

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "Controller Mode",
        after_update_cb: RVCallbackType[HVACControllerMode] | None = None,
    ) -> None:
        """Initialize remote value of KNX climate mode."""
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    def supported_operation_modes(self) -> list[HVACOperationMode]:
        """Return a list of all supported operation modes."""
        return []

    def set_operation_mode(self, mode: HVACOperationMode) -> None:
        """Set operation mode. Return if not supported."""
        return

    def supported_controller_modes(self) -> list[HVACControllerMode]:
        """Return a list of all supported controller modes."""
        return DPTHVACContrMode.get_valid_values()

    def set_controller_mode(self, mode: HVACControllerMode) -> None:
        """Set controller mode. Return if not supported."""
        if mode not in self.supported_controller_modes():
            return
        super().set(mode)


class RemoteValueHVACStatus(RemoteValueClimateModeBase[HVACStatus]):
    """Abstraction for remote value of KNX climate HVAC status (Eberle status)."""

    dpt_class = DPTHVACStatus

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "Controller status",
        after_update_cb: RVCallbackType[HVACStatus] | None = None,
    ) -> None:
        """Initialize remote value of KNX climate controller status."""
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    def supported_operation_modes(self) -> list[HVACOperationMode]:
        """Return a list of all supported operation modes."""
        return [
            HVACOperationMode.COMFORT,
            HVACOperationMode.STANDBY,
            HVACOperationMode.ECONOMY,
            HVACOperationMode.BUILDING_PROTECTION,
        ]

    def set_operation_mode(self, mode: HVACOperationMode) -> None:
        """Set operation mode. Return if not supported."""
        if mode not in self.supported_operation_modes():
            return
        if self._value is None:
            raise ConversionError(
                "HVACStatus value not initialized. Can not write new mode.",
                device_name=self.device_name,
            )
        new_status = HVACStatus(
            mode=mode,
            dew_point=self._value.dew_point,
            heat_cool=self._value.heat_cool,
            inactive=self._value.inactive,
            frost_alarm=self._value.frost_alarm,
        )
        return super().set(new_status)

    def supported_controller_modes(self) -> list[HVACControllerMode]:
        """Return a list of all supported controller modes."""
        return [HVACControllerMode.HEAT, HVACControllerMode.COOL]

    def set_controller_mode(self, mode: HVACControllerMode) -> None:
        """Set controller mode. Return if not supported."""
        if mode not in self.supported_controller_modes():
            return
        if self._value is None:
            raise ConversionError(
                "HVACStatus value not initialized. Can not write new mode.",
                device_name=self.device_name,
            )
        new_status = HVACStatus(
            mode=self._value.mode,
            dew_point=self._value.dew_point,
            heat_cool=HeatCool.HEAT
            if mode == HVACControllerMode.HEAT
            else HeatCool.COOL,
            inactive=self._value.inactive,
            frost_alarm=self._value.frost_alarm,
        )
        super().set(new_status)


class RemoteValueBinaryOperationMode(
    RemoteValueClimateModeBase[HVACOperationMode | None]
):
    """Abstraction for remote value of split up KNX climate modes."""

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "Climate mode binary",
        after_update_cb: RVCallbackType[HVACOperationMode | None] | None = None,
        operation_mode: HVACOperationMode | None = None,
    ) -> None:
        """Initialize remote value of KNX DPT 1 representing a climate operation mode."""
        if not isinstance(operation_mode, HVACOperationMode):
            raise ConversionError(
                "Invalid operation mode type",
                operation_mode=str(operation_mode),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        self.operation_mode = operation_mode
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    def supported_operation_modes(self) -> list[HVACOperationMode]:
        """Return a list of all supported operation modes."""
        # all binary operation modes off -> Standby according to MDT
        return (
            [self.operation_mode, HVACOperationMode.STANDBY]
            if self.operation_mode is not HVACOperationMode.STANDBY
            else [HVACOperationMode.STANDBY]
        )

    def set_operation_mode(self, mode: HVACOperationMode) -> None:
        """Set operation mode. Return if not supported."""
        super().set(mode)

    def supported_controller_modes(self) -> list[HVACControllerMode]:
        """Return a list of all supported controller modes."""
        return []

    def set_controller_mode(self, mode: HVACControllerMode) -> None:
        """Set controller mode. Return if not supported."""
        return

    def to_knx(self, value: HVACOperationMode | None) -> DPTBinary:
        """Convert value to payload."""
        if isinstance(value, HVACOperationMode):
            # foreign operation modes will set the RemoteValue to False
            return DPTBinary(value == self.operation_mode)
        raise ConversionError(
            "value invalid",
            value=value,
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    def from_knx(self, payload: DPTArray | DPTBinary) -> HVACOperationMode | None:
        """Convert current payload to value."""
        if payload.value == 1:
            return self.operation_mode
        if payload.value == 0:
            return None
        raise CouldNotParseTelegram(
            "Payload invalid",
            payload=str(payload),
            device_name=self.device_name,
            feature_name=self.feature_name,
        )


class RemoteValueBinaryHeatCool(RemoteValueClimateModeBase[HVACControllerMode]):
    """Abstraction for remote value of heat/cool controller mode."""

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "Controller mode Heat/Cool",
        after_update_cb: RVCallbackType[HVACControllerMode] | None = None,
        controller_mode: HVACControllerMode | None = None,
    ) -> None:
        """Initialize remote value of KNX DPT 1 representing a climate controller mode."""
        if not isinstance(controller_mode, HVACControllerMode):
            raise ConversionError(
                "Invalid controller mode type",
                controller_mode=str(controller_mode),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        if controller_mode not in self.supported_controller_modes():
            raise ConversionError(
                "Controller mode not supported for binary mode object",
                controller_mode=str(controller_mode),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        self.controller_mode = controller_mode
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    def supported_operation_modes(self) -> list[HVACOperationMode]:
        """Return a list of all supported operation modes."""
        return []

    def set_operation_mode(self, mode: HVACOperationMode) -> None:
        """Set operation mode. Return if not supported."""
        return

    def supported_controller_modes(self) -> list[HVACControllerMode]:
        """Return a list of all supported controller modes."""
        return [HVACControllerMode.HEAT, HVACControllerMode.COOL]

    def set_controller_mode(self, mode: HVACControllerMode) -> None:
        """Set controller mode. Return if not supported."""
        if mode not in self.supported_controller_modes():
            return
        super().set(mode)

    def to_knx(self, value: Any) -> DPTBinary:
        """Convert value to payload."""
        if isinstance(value, HVACControllerMode):
            # foreign operation modes will set the RemoteValue to False
            return DPTBinary(value == self.controller_mode)
        raise ConversionError(
            "value invalid",
            value=value,
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    def from_knx(self, payload: DPTArray | DPTBinary) -> HVACControllerMode:
        """Convert current payload to value."""
        if payload.value == 1:
            return self.controller_mode
        if payload.value == 0:
            # return the other operation mode
            return next(
                _op
                for _op in self.supported_controller_modes()
                if _op is not self.controller_mode
            )
        raise CouldNotParseTelegram(
            "Payload invalid",
            payload=str(payload),
            device_name=self.device_name,
            feature_name=self.feature_name,
        )



"""Unit test for KNX/IP DeviceConfigurationAck objects."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import DeviceConfigurationAck, ErrorCode, KNXIPFrame


class TestKNXIPDeviceConfigurationAck:
    """Test class for KNX/IP DeviceConfigurationAck objects."""

    def test_device_configuration_ack(self) -> None:
        """Test parsing and streaming device configuration ACK KNX/IP packet."""
        raw = bytes((0x06, 0x10, 0x03, 0x11, 0x00, 0x0A, 0x04, 0x2A, 0x17, 0x00))
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, DeviceConfigurationAck)
        assert knxipframe.body.communication_channel_id == 42
        assert knxipframe.body.sequence_counter == 23
        assert knxipframe.body.status_code == ErrorCode.E_NO_ERROR

        device_configuration_ack = DeviceConfigurationAck(
            communication_channel_id=42,
            sequence_counter=23,
        )
        knxipframe2 = KNXIPFrame.init_from_body(device_configuration_ack)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_ack_information(self) -> None:
        """Test parsing and streaming wrong DeviceConfigurationAck (wrong length byte)."""
        raw = bytes((0x06, 0x10, 0x03, 0x11, 0x00, 0x0A, 0x03, 0x2A, 0x17, 0x00))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_ack_information2(self) -> None:
        """Test parsing and streaming wrong DeviceConfigurationAck (wrong length)."""
        raw = bytes((0x06, 0x10, 0x03, 0x11, 0x00, 0x0A, 0x04, 0x2A, 0x17))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""Test management procedures."""

import asyncio
from unittest.mock import AsyncMock, call

import pytest

from xknx import XKNX
from xknx.exceptions import ManagementConnectionError
from xknx.management import procedures
from xknx.management.management import MANAGAMENT_CONNECTION_TIMEOUT
from xknx.telegram import (
    GroupAddress,
    IndividualAddress,
    Telegram,
    TelegramDirection,
    apci,
    tpci,
)

from ..conftest import EventLoopClockAdvancer


async def test_dm_restart() -> None:
    """Test dm_restart."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address = IndividualAddress("4.0.10")

    connect = Telegram(destination_address=individual_address, tpci=tpci.TConnect())
    restart = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDataConnected(0),
        payload=apci.Restart(),
    )
    disconnect = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDisconnect(),
    )
    await procedures.dm_restart(xknx, individual_address)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(restart),
        call(disconnect),
    ]


async def test_nm_individual_address_check_success() -> None:
    """Test nm_individual_address_check."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address = IndividualAddress("4.0.10")

    connect = Telegram(destination_address=individual_address, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    ack = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TAck(0),
    )
    device_desc_resp = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorResponse(),
    )
    task = asyncio.create_task(
        procedures.nm_individual_address_check(xknx, individual_address)
    )
    await asyncio.sleep(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
    ]
    # receive response
    xknx.management.process(ack)
    xknx.management.process(device_desc_resp)
    assert await task


async def test_nm_individual_address_check_refused() -> None:
    """Test nm_individual_address_check."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address = IndividualAddress("4.0.10")

    connect = Telegram(destination_address=individual_address, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    ack = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TAck(0),
    )
    disconnect = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TDisconnect(),
    )
    task = asyncio.create_task(
        procedures.nm_individual_address_check(xknx, individual_address)
    )
    await asyncio.sleep(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
    ]
    xknx.management.process(disconnect)
    xknx.management.process(ack)
    assert await task


async def test_nm_individual_address_read(time_travel: EventLoopClockAdvancer) -> None:
    """Test nm_individual_address_read."""
    _timeout = 2
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address_1 = IndividualAddress("1.1.4")
    individual_address_2 = IndividualAddress("15.15.255")

    task = asyncio.create_task(
        procedures.nm_individual_address_read(xknx=xknx, timeout=_timeout)
    )
    address_broadcast = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )

    address_reply_message_1 = Telegram(
        source_address=individual_address_1,
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressResponse(),
    )

    address_reply_message_2 = Telegram(
        source_address=individual_address_2,
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressResponse(),
    )

    await asyncio.sleep(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(address_broadcast),
    ]
    xknx.management.process(address_reply_message_1)
    xknx.management.process(address_reply_message_2)
    await time_travel(_timeout)
    assert await task


async def test_nm_individual_address_read_multiple() -> None:
    """Test nm_individual_address_read."""
    _timeout = 2
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address_1 = IndividualAddress("1.1.4")
    individual_address_2 = IndividualAddress("15.15.255")

    task = asyncio.create_task(
        procedures.nm_individual_address_read(
            xknx=xknx, timeout=_timeout, raise_if_multiple=True
        )
    )
    address_broadcast = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )

    address_reply_message_1 = Telegram(
        source_address=individual_address_1,
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressResponse(),
    )

    address_reply_message_2 = Telegram(
        source_address=individual_address_2,
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressResponse(),
    )

    await asyncio.sleep(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(address_broadcast),
    ]
    xknx.management.process(address_reply_message_1)
    xknx.management.process(address_reply_message_2)
    # no need to wait for _timeout due to `raise_if_multiple=True``
    with pytest.raises(ManagementConnectionError):
        await task


async def test_nm_individual_address_write(time_travel: EventLoopClockAdvancer) -> None:
    """Test nm_individual_address_write."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address_old = IndividualAddress("15.15.255")
    individual_address_new = IndividualAddress("1.1.4")

    connect = Telegram(destination_address=individual_address_new, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDataConnected(0),
        direction=TelegramDirection.OUTGOING,
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )

    address_reply_message = Telegram(
        source_address=individual_address_old,
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.INCOMING,
        payload=apci.IndividualAddressResponse(),
    )

    device_desc_resp = Telegram(
        source_address=individual_address_new,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorResponse(),
    )
    ack = Telegram(
        source_address=individual_address_new,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TAck(0),
    )
    ack2 = Telegram(
        destination_address=individual_address_new,
        source_address=IndividualAddress(0),
        direction=TelegramDirection.OUTGOING,
        tpci=tpci.TAck(0),
    )
    disconnect = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDisconnect(),
    )
    individual_address_read = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )
    individual_address_write = Telegram(
        GroupAddress("0/0/0"),
        payload=apci.IndividualAddressWrite(address=individual_address_new),
    )
    task = asyncio.create_task(
        procedures.nm_individual_address_write(
            xknx=xknx, individual_address=individual_address_new
        )
    )

    # make sure first request (address check) times out
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)

    # send response to device in programming mode
    xknx.management.process(address_reply_message)

    # confirm device is up and running
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)
    xknx.management.process(ack)
    xknx.management.process(device_desc_resp)

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
        call(device_desc_read),  # due to retransmit
        call(disconnect),
        call(individual_address_read),
        call(individual_address_write),
        call(connect),
        call(device_desc_read),
        call(ack2),
    ]

    await task


async def test_nm_individual_address_write_two_devices_in_programming_mode(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_write."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address_old = IndividualAddress("15.15.255")
    individual_address_new = IndividualAddress("1.1.4")

    connect = Telegram(destination_address=individual_address_new, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    address_reply_message = Telegram(
        source_address=individual_address_old,
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.INCOMING,
        payload=apci.IndividualAddressResponse(),
    )
    disconnect = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDisconnect(),
    )
    individual_address_read = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )

    task = asyncio.create_task(
        procedures.nm_individual_address_write(
            xknx=xknx, individual_address=individual_address_new
        )
    )

    # make sure first request (address check) times out
    await time_travel(0)  # start
    await time_travel(3)  # first timeout
    await time_travel(3)  # second timeout

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
        call(device_desc_read),  # due to retransmit
        call(disconnect),
        call(individual_address_read),
    ]
    # receive two responses from devices in programming mode
    xknx.management.process(address_reply_message)
    xknx.management.process(address_reply_message)
    with pytest.raises(
        ManagementConnectionError,
        match="More than one KNX device is in programming mode",
    ):
        await task
    assert len(xknx.cemi_handler.send_telegram.call_args_list) == 5


async def test_nm_individual_address_write_no_device_programming_mode(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_write."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address_new = IndividualAddress("1.1.4")

    connect = Telegram(destination_address=individual_address_new, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    disconnect = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDisconnect(),
    )
    individual_address_read = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )

    task = asyncio.create_task(
        procedures.nm_individual_address_write(
            xknx=xknx, individual_address=individual_address_new
        )
    )

    # make sure first request (address check) times out
    await time_travel(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
    ]
    # first timeout - retransmit DeviceDescriptorRead
    await time_travel(3)
    assert xknx.cemi_handler.send_telegram.call_args_list[2:] == [
        call(device_desc_read),
    ]
    # retry also timed out
    await time_travel(3)
    assert xknx.cemi_handler.send_telegram.call_args_list[3:] == [
        call(disconnect),
        call(individual_address_read),
    ]
    # IndividualAddressRead also times out
    await time_travel(3)
    with pytest.raises(
        ManagementConnectionError, match="No device in programming mode"
    ):
        await task
    assert len(xknx.cemi_handler.send_telegram.call_args_list) == 5


async def test_nm_individual_address_write_address_found(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_write."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address = IndividualAddress("1.1.4")

    connect = Telegram(destination_address=individual_address, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    ack_in = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TAck(0),
    )
    ack_out = Telegram(
        source_address=IndividualAddress(0),
        destination_address=individual_address,
        tpci=tpci.TAck(0),
    )
    device_desc_resp = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorResponse(),
    )
    disconnect = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDisconnect(),
    )
    individual_address_read = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )

    task = asyncio.create_task(
        procedures.nm_individual_address_write(
            xknx=xknx, individual_address=individual_address
        )
    )

    # first request (address check) succeeds
    await time_travel(0)
    xknx.management.process(ack_in)
    xknx.management.process(device_desc_resp)

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
        call(ack_out),
    ]
    await time_travel(0)
    assert xknx.cemi_handler.send_telegram.call_args_list[3:] == [
        call(disconnect),
        call(individual_address_read),
    ]
    # second request times out - no device in programming mode
    await time_travel(3)
    with pytest.raises(
        ManagementConnectionError, match="No device in programming mode"
    ):
        await task
    assert len(xknx.cemi_handler.send_telegram.call_args_list) == 5


async def test_nm_individual_address_write_programming_failed(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_write."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address_old = IndividualAddress("15.15.255")
    individual_address_new = IndividualAddress("1.1.4")

    connect = Telegram(destination_address=individual_address_new, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDataConnected(0),
        direction=TelegramDirection.OUTGOING,
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )

    address_reply_message = Telegram(
        source_address=individual_address_old,
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.INCOMING,
        payload=apci.IndividualAddressResponse(),
    )
    disconnect = Telegram(
        destination_address=individual_address_new,
        tpci=tpci.TDisconnect(),
    )
    individual_address_read = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )
    individual_address_write = Telegram(
        GroupAddress("0/0/0"),
        payload=apci.IndividualAddressWrite(address=individual_address_new),
    )
    task = asyncio.create_task(
        procedures.nm_individual_address_write(
            xknx=xknx, individual_address=individual_address_new
        )
    )

    # make sure first request (address check) times out
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)

    # send response to device in programming mode
    xknx.management.process(address_reply_message)

    # device experienced error, so set connection request timeout
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
        call(device_desc_read),  # due to retransmit
        call(disconnect),
        call(individual_address_read),
        call(individual_address_write),
        call(connect),
        call(device_desc_read),
        call(device_desc_read),
        call(disconnect),
    ]

    with pytest.raises(ManagementConnectionError):
        await task


async def test_nm_individual_address_write_address_found_other_in_programming_mode(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_write."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address = IndividualAddress("1.1.5")
    individual_address_pgm = IndividualAddress("1.1.4")

    connect = Telegram(destination_address=individual_address, tpci=tpci.TConnect())
    device_desc_read = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    ack = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TAck(0),
    )
    ack2 = Telegram(
        source_address=IndividualAddress(0),
        destination_address=individual_address,
        tpci=tpci.TAck(0),
    )
    device_desc_resp = Telegram(
        source_address=individual_address,
        destination_address=IndividualAddress(0),
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorResponse(),
    )
    disconnect = Telegram(
        destination_address=individual_address,
        tpci=tpci.TDisconnect(),
    )
    individual_address_read = Telegram(
        GroupAddress("0/0/0"), payload=apci.IndividualAddressRead()
    )
    address_reply_message = Telegram(
        source_address=individual_address_pgm,
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.INCOMING,
        payload=apci.IndividualAddressResponse(),
    )

    task = asyncio.create_task(
        procedures.nm_individual_address_write(
            xknx=xknx, individual_address=individual_address
        )
    )

    # make sure first request (address check) times out
    await time_travel(0)
    xknx.management.process(ack)
    xknx.management.process(device_desc_resp)
    await time_travel(MANAGAMENT_CONNECTION_TIMEOUT)

    xknx.management.process(address_reply_message)

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(connect),
        call(device_desc_read),
        call(ack2),
        call(disconnect),
        call(individual_address_read),
    ]

    with pytest.raises(ManagementConnectionError):
        await task


async def test_nm_individual_address_serial_number_read(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_serial_number_read."""

    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    individual_address = IndividualAddress("1.1.5")
    serial_number = b"aabbccddeeff"

    task = asyncio.create_task(
        procedures.nm_individual_address_serial_number_read(
            xknx=xknx, serial=serial_number
        )
    )

    read_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialRead(serial=serial_number),
    )
    address_reply = Telegram(
        source_address=individual_address,
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.INCOMING,
        payload=apci.IndividualAddressSerialResponse(
            address=individual_address, serial=serial_number
        ),
    )

    await time_travel(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(read_address),
    ]
    xknx.management.process(address_reply)

    assert await task == individual_address


async def test_nm_individual_address_serial_number_read_fail(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_serial_number_read."""

    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    serial_number = b"aabbccddeeff"

    task = asyncio.create_task(
        procedures.nm_individual_address_serial_number_read(
            xknx=xknx, serial=serial_number
        )
    )

    read_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialRead(serial=serial_number),
    )

    await time_travel(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(read_address),
    ]
    await time_travel(3)

    assert await task is None


async def test_nm_individual_address_serial_number_write(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_serial_number_write."""

    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    serial_number = b"aabbccddeeff"
    individual_address = IndividualAddress("1.1.5")

    task = asyncio.create_task(
        procedures.nm_individual_address_serial_number_write(
            xknx=xknx, serial=serial_number, individual_address=individual_address
        )
    )

    write_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialWrite(
            serial=serial_number, address=individual_address
        ),
    )
    read_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialRead(serial=serial_number),
    )
    address_reply = Telegram(
        source_address=individual_address,
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.INCOMING,
        payload=apci.IndividualAddressSerialResponse(
            address=individual_address, serial=serial_number
        ),
    )

    await time_travel(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(write_address),
        call(read_address),
    ]
    xknx.management.process(address_reply)
    await task


async def test_nm_individual_address_serial_number_write_fail_no_response(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_serial_number_write."""

    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    serial_number = b"aabbccddeeff"
    individual_address = IndividualAddress("1.1.5")

    task = asyncio.create_task(
        procedures.nm_individual_address_serial_number_write(
            xknx=xknx, serial=serial_number, individual_address=individual_address
        )
    )

    write_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialWrite(
            serial=serial_number, address=individual_address
        ),
    )
    read_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialRead(serial=serial_number),
    )

    await time_travel(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(write_address),
        call(read_address),
    ]
    await time_travel(3)
    with pytest.raises(ManagementConnectionError):
        await task


async def test_nm_individual_address_serial_number_write_fail_wrong_address(
    time_travel: EventLoopClockAdvancer,
) -> None:
    """Test nm_individual_address_serial_number_write."""

    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    serial_number = b"aabbccddeeff"
    individual_address_tx = IndividualAddress("1.1.5")
    individual_address_rx = IndividualAddress("1.1.6")

    task = asyncio.create_task(
        procedures.nm_individual_address_serial_number_write(
            xknx=xknx, serial=serial_number, individual_address=individual_address_tx
        )
    )

    write_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialWrite(
            serial=serial_number, address=individual_address_tx
        ),
    )
    read_address = Telegram(
        destination_address=GroupAddress("0/0/0"),
        payload=apci.IndividualAddressSerialRead(serial=serial_number),
    )
    address_reply = Telegram(
        source_address=individual_address_rx,
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.INCOMING,
        payload=apci.IndividualAddressSerialResponse(
            address=individual_address_rx, serial=serial_number
        ),
    )

    await time_travel(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(write_address),
        call(read_address),
    ]
    xknx.management.process(address_reply)
    with pytest.raises(ManagementConnectionError):
        await task



"""
Module for Serialization and Deserialization of KNX RoutingIndication frames.

Routing indications are used to transport CEMI Messages.
"""

from __future__ import annotations

from .body import KNXIPBody
from .knxip_enum import KNXIPServiceType


class RoutingIndication(KNXIPBody):
    """Representation of a KNX RoutingIndication frame."""

    SERVICE_TYPE = KNXIPServiceType.ROUTING_INDICATION

    def __init__(self, raw_cemi: bytes = b"") -> None:
        """Initialize RoutingIndication object."""
        self.raw_cemi = raw_cemi

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return len(self.raw_cemi)

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        self.raw_cemi = raw
        return len(raw)

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return self.raw_cemi

    def __repr__(self) -> str:
        """Return object as readable string."""
        return f'<RoutingIndication cemi="{self.raw_cemi.hex()}" />'



"""
Module for Serialization and Deserialization of KNX Session Responses.

The SESSION_RESPONSE shall be sent by the KNXnet/IP secure server to the secure
client's control endpoint in response to a received secure session request frame.
"""

from __future__ import annotations

from typing import Final

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .knxip_enum import KNXIPServiceType


class SessionResponse(KNXIPBody):
    """Representation of a KNX Session Response."""

    SERVICE_TYPE = KNXIPServiceType.SESSION_RESPONSE
    # 2 octets secure session identifier
    # 32 octets for the servers ECDH public value
    # 16 octets for the message authentication code
    LENGTH: Final = 50

    def __init__(
        self,
        secure_session_id: int = 0,
        ecdh_server_public_key: bytes = bytes(32),
        message_authentication_code: bytes = bytes(16),
    ) -> None:
        """Initialize SessionResponse object."""
        self.ecdh_server_public_key = ecdh_server_public_key
        # secure session identifier 0 shall in general be reserved for
        # multicast data and shall not be used for unicast connections
        self.secure_session_id = secure_session_id
        self.message_authentication_code = message_authentication_code

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return SessionResponse.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) != SessionResponse.LENGTH:
            raise CouldNotParseKNXIP("SessionResponse has wrong length")
        self.secure_session_id = int.from_bytes(raw[:2], "big")
        self.ecdh_server_public_key = raw[2:34]
        self.message_authentication_code = raw[34:]
        return SessionResponse.LENGTH

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            self.secure_session_id.to_bytes(2, "big")
            + self.ecdh_server_public_key
            + self.message_authentication_code
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            f"<SessionResponse "
            f'secure_session_id="{self.secure_session_id}" '
            f'ecdh_server_public_key="{self.ecdh_server_public_key.hex()}" '
            f'message_authentication_code="{self.message_authentication_code.hex()}" />'
        )



"""Unit test for KNX/IP DescriptionResponse objects."""

from xknx.knxip import (
    DescriptionResponse,
    DIBDeviceInformation,
    DIBServiceFamily,
    DIBSuppSVCFamilies,
    KNXIPFrame,
)


class TestKNXIPDescriptionResponse:
    """Test class for KNX/IP DescriptionResponse objects."""

    def test_description_response(self) -> None:
        """Test parsing and streaming DescriptionResponse KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 02 04 00 48 36 01 02 00 10 00 00 00 00 08"
            "2d 40 83 4d e0 00 17 0c 00 0a b3 27 4a 32 4b 4e"
            "58 2f 49 50 2d 52 6f 75 74 65 72 00 00 00 00 00"
            "00 00 00 00 00 00 00 00 00 00 00 00 0c 02 02 02"
            "03 02 04 02 05 02 07 01",
        )
        knxipframe, rest = KNXIPFrame.from_knx(raw)
        assert not rest
        assert knxipframe.header.total_length == 72
        assert knxipframe.to_knx() == raw

        assert isinstance(knxipframe.body, DescriptionResponse)
        assert len(knxipframe.body.dibs) == 2
        # Specific testing of parsing and serializing of
        # DIBDeviceInformation and DIBSuppSVCFamilies is
        # done within knxip_dib_test.py
        assert isinstance(knxipframe.body.dibs[0], DIBDeviceInformation)
        assert isinstance(knxipframe.body.dibs[1], DIBSuppSVCFamilies)
        assert knxipframe.body.device_name == "KNX/IP-Router"
        assert knxipframe.body.dibs[1].supports(DIBServiceFamily.CORE, version=2)
        assert knxipframe.body.dibs[1].supports(
            DIBServiceFamily.DEVICE_MANAGEMENT, version=2
        )
        assert knxipframe.body.dibs[1].supports(DIBServiceFamily.ROUTING, version=2)
        assert knxipframe.body.dibs[1].supports(DIBServiceFamily.TUNNELING, version=2)
        assert knxipframe.body.dibs[1].supports(
            DIBServiceFamily.REMOTE_CONFIGURATION_DIAGNOSIS
        )
        assert not knxipframe.body.dibs[1].supports(DIBServiceFamily.OBJECT_SERVER)

        description_response = DescriptionResponse()
        description_response.dibs.append(knxipframe.body.dibs[0])
        description_response.dibs.append(knxipframe.body.dibs[1])
        knxipframe2 = KNXIPFrame.init_from_body(description_response)

        assert knxipframe2.to_knx() == raw

    def test_unknown_device_name(self) -> None:
        """Test device_name if no DIBDeviceInformation is present."""
        description_response = DescriptionResponse()
        assert description_response.device_name == "UNKNOWN"



"""Conftest for XKNX."""

import asyncio
import sys
from unittest.mock import AsyncMock, Mock, patch

import pytest

from xknx import XKNX


class EventLoopClockAdvancer:
    """Allow advancing of loop time."""

    # thanks to @dermotduffy for his asyncio.sleep mock
    # https://github.com/dermotduffy/hyperion-py/blob/main/tests/client_test.py#L273

    __slots__ = ("_base_time", "loop", "offset")

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        """Initialize."""
        self.offset = 0.0
        self._base_time = loop.time
        self.loop = loop

        # incorporate offset timing into the event loop
        self.loop.time = self.time  # type: ignore[method-assign]

    def time(self) -> float:
        """Return loop time adjusted by offset."""
        return self._base_time() + self.offset

    async def _exhaust_callbacks(self) -> None:
        """Run the loop until all ready callbacks are executed."""
        while self.loop._ready:  # type: ignore[attr-defined] # noqa: ASYNC110
            await asyncio.sleep(0)

    async def __call__(self, seconds: float) -> None:
        """Advance time by a given offset in seconds."""
        # Exhaust all callbacks.
        await self._exhaust_callbacks()

        if seconds > 0:
            # advance the clock by the given offset
            self.offset += seconds
            # Once the clock is adjusted, new tasks may have just been
            # scheduled for running in the next pass through the event loop
            await asyncio.sleep(0)
            await self._exhaust_callbacks()


@pytest.fixture
async def time_travel() -> EventLoopClockAdvancer:
    """Advance loop time and run callbacks."""
    event_loop = asyncio.get_running_loop()
    return EventLoopClockAdvancer(event_loop)


@pytest.fixture
def xknx_no_interface() -> XKNX:
    """Return XKNX instance without KNX/IP interface."""

    def knx_ip_interface_mock() -> Mock:
        """Create a xknx knx ip interface mock."""
        mock = Mock()
        mock.start = AsyncMock()
        mock.stop = AsyncMock()
        mock.send_cemi = AsyncMock()
        return mock

    with patch("xknx.xknx.knx_interface_factory", return_value=knx_ip_interface_mock()):
        return XKNX()


# py3.10 doesn't properly support patch() with wraps and autospec https://github.com/python/cpython/pull/117124
skip_3_10 = pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason="requires Python 3.11 or higher",
)



"""Example on how to connect to restart a KNX device."""

import asyncio
import sys

from xknx import XKNX
from xknx.management.procedures import dm_restart
from xknx.telegram import IndividualAddress


async def main(argv: list[str]) -> int:
    """Restart a KNX device."""
    if len(argv) != 2:
        print(f"{argv[0]}: missing target address.")
        return 1

    address = IndividualAddress(argv[1])

    async with XKNX() as xknx:
        await dm_restart(xknx, address)
    return 0


if __name__ == "__main__":
    asyncio.run(main(sys.argv))



"""Unit test for KNX binary/integer objects."""

from inspect import isabstract
from typing import Any

import pytest

from xknx.dpt import (
    DPT2ByteFloat,
    DPT2ByteUnsigned,
    DPTActiveEnergy,
    DPTArray,
    DPTBase,
    DPTBinary,
    DPTColorRGBW,
    DPTComplex,
    DPTConsumerProducer,
    DPTEnum,
    DPTHVACContrMode,
    DPTNumeric,
    DPTScaling,
    DPTString,
    DPTTemperature,
)
from xknx.exceptions import CouldNotParseTelegram


class TestDPTBase:
    """Test class for transcoder base object."""

    def test_dpt_abstract_subclasses_ignored(self) -> None:
        """Test if abstract base classes are ignored by dpt_class_tree and __recursive_subclasses__."""
        for dpt in DPTBase.dpt_class_tree():
            assert dpt not in (DPTBase, DPTNumeric, DPTEnum, DPTComplex)
            assert not isabstract(dpt)
            assert dpt()  # test for abstract class - to be removed if instantiation is not allowed anymore

    def test_dpt_concrete_subclasses_included(self) -> None:
        """Test if concrete subclasses are included by dpt_class_tree."""
        for dpt in (
            DPT2ByteFloat,
            DPTString,
            DPTTemperature,
            DPTScaling,
            DPTHVACContrMode,
            DPTColorRGBW,
            DPTConsumerProducer,
        ):
            assert dpt in DPTBase.dpt_class_tree()

    @pytest.mark.parametrize("dpt_class", [DPTString, DPT2ByteFloat])
    def test_dpt_non_abstract_baseclass_included(
        self, dpt_class: type[DPTBase]
    ) -> None:
        """Test if non-abstract base classes is included by dpt_class_tree."""
        assert dpt_class in dpt_class.dpt_class_tree()

    def test_dpt_subclasses_definition_types(self) -> None:
        """Test value_type and dpt_*_number values for correct type in subclasses of DPTBase."""
        for dpt in DPTBase.dpt_class_tree():
            assert isinstance(dpt.value_type, str), (
                f"Wrong type for value_type in {dpt} : {type(dpt.value_type)} - str expected"
            )
            assert dpt.value_type, f"Empty string for value_type in {dpt} not allowed"

            assert isinstance(dpt.dpt_main_number, int), (
                f"Wrong type for dpt_main_number in {dpt} : {type(dpt.dpt_main_number)} - int expected"
            )
            assert dpt.dpt_main_number, (
                f"Zero value for dpt_main_number in {dpt} not allowed"
            )

            assert isinstance(dpt.dpt_sub_number, int | type(None)), (
                f"Wrong type for dpt_sub_number in {dpt} : {type(dpt.dpt_sub_number)} - int or `None` expected"
            )

    def test_dpt_subclasses_no_duplicate_value_types(self) -> None:
        """Test for duplicate value_type values in subclasses of DPTBase."""
        value_types = [
            dpt.value_type
            for dpt in DPTBase.dpt_class_tree()
            if dpt.value_type is not None
        ]
        assert len(value_types) == len(set(value_types)), (
            f"Duplicate DPT value_types found: { {item for item in value_types if value_types.count(item) > 1} }"
        )

    def test_dpt_subclasses_no_duplicate_dpt_number(self) -> None:
        """Test for duplicate value_type values in subclasses of DPTBase."""
        dpt_tuples = [
            (dpt.dpt_main_number, dpt.dpt_sub_number)
            for dpt in DPTBase.dpt_class_tree()
        ]
        assert len(dpt_tuples) == len(set(dpt_tuples)), (
            f"Duplicate DPT numbers found: { {item for item in dpt_tuples if dpt_tuples.count(item) > 1} }"
        )

    @pytest.mark.parametrize(
        "equal_dpts",
        [
            # strings in dictionaries would fail type checking, but should work nevertheless
            ["2byte_unsigned", 7, "DPT-7", {"main": 7}, {"main": "7", "sub": None}],
            ["temperature", "9.001", {"main": 9, "sub": 1}, {"main": "9", "sub": "1"}],
            ["active_energy", "13.010", {"main": 13, "sub": 10}],
            ["consumer_producer", "1.1200", {"main": 1, "sub": 1200}],
        ],
    )
    def test_dpt_alternative_notations(self, equal_dpts: list[Any]) -> None:
        """Test the parser for accepting alternative notations for the same DPT class."""
        parsed = [DPTBase.parse_transcoder(dpt) for dpt in equal_dpts]
        assert issubclass(parsed[0], DPTBase)
        assert all(parsed[0] == dpt for dpt in parsed)

    @pytest.mark.parametrize(
        "equal_dpts",
        [
            # strings in dictionaries would fail type checking, but should work nevertheless
            [
                "2byte_unsigned",
                7,
                "DPT-7",
                {"main": 7},
                {"main": "7", "sub": None},
                DPT2ByteUnsigned,
            ],
            [
                "temperature",
                "9.001",
                {"main": 9, "sub": 1},
                {"main": "9", "sub": "1"},
                DPTTemperature,
            ],
            ["active_energy", "13.010", {"main": 13, "sub": 10}, DPTActiveEnergy],
        ],
    )
    def test_get_dpt_alternative_notations(self, equal_dpts: list[Any]) -> None:
        """Test the parser for accepting alternative notations for the same DPT class."""
        parsed = [DPTBase.get_dpt(dpt) for dpt in equal_dpts]
        assert issubclass(parsed[0], DPTBase)
        assert all(parsed[0] == dpt for dpt in parsed)

    INVALID_DPT_IDENTIFIERS = [
        None,
        0,
        999999999,
        9.001,  # float is not valid
        "invalid_string",
        {"sub": 1},
        {"main": None, "sub": None},
        {"main": "invalid"},
        {"main": 9, "sub": "invalid"},
        [9, 1],
        (9,),
    ]

    @pytest.mark.parametrize("value", INVALID_DPT_IDENTIFIERS)
    def test_parse_transcoder_invalid_data(self, value: Any) -> None:
        """Test parsing invalid data."""
        assert DPTBase.parse_transcoder(value) is None

    @pytest.mark.parametrize("value", INVALID_DPT_IDENTIFIERS)
    def test_get_dpt_invalid_data(self, value: Any) -> None:
        """Test parsing invalid data."""
        with pytest.raises(ValueError):
            DPTBase.get_dpt(value)

    def test_parse_transcoder_from_subclass(self) -> None:
        """Test parsing only subclasses of a DPT class."""
        assert DPTBase.parse_transcoder("string") == DPTString
        assert DPTNumeric.parse_transcoder("string") is None
        assert DPT2ByteFloat.parse_transcoder("string") is None

        assert DPTBase.parse_transcoder("percent") == DPTScaling
        assert DPTNumeric.parse_transcoder("percent") == DPTScaling
        assert DPT2ByteFloat.parse_transcoder("percent") is None

        assert DPTBase.parse_transcoder("temperature") == DPTTemperature
        assert DPTNumeric.parse_transcoder("temperature") == DPTTemperature
        assert DPT2ByteFloat.parse_transcoder("temperature") == DPTTemperature

    def test_get_dpt_from_subclass(self) -> None:
        """Test parsing only subclasses of a DPT class."""
        assert DPTBase.get_dpt("string") == DPTString
        with pytest.raises(ValueError):
            DPTNumeric.get_dpt("string")

        assert DPTBase.get_dpt("percent") == DPTScaling
        assert DPTNumeric.get_dpt("percent") == DPTScaling

        assert DPTBase.get_dpt("temperature") == DPTTemperature
        assert DPTNumeric.get_dpt("temperature") == DPTTemperature
        assert DPT2ByteFloat.get_dpt("temperature") == DPTTemperature

    def test_dpt_name(self) -> None:
        """Test DPT name."""
        assert DPTBase.dpt_name() == "DPTBase (abstract)"
        assert DPTNumeric.dpt_name() == "DPTNumeric (abstract)"
        assert DPT2ByteFloat.dpt_name() == "DPT2ByteFloat (9)"
        assert DPTString.dpt_name() == "DPTString (16.000)"
        assert DPTColorRGBW.dpt_name() == "DPTColorRGBW (251.600)"
        assert DPTConsumerProducer.dpt_name() == "DPTConsumerProducer (1.1200)"


class TestDPTBaseSubclass:
    """Test subclass of transcoder base object."""

    @pytest.mark.parametrize("dpt_class", DPTBase.dpt_class_tree())
    def test_required_values(self, dpt_class: type[DPTBase]) -> None:
        """Test required class variables are set for definitions."""
        assert dpt_class.payload_type in (DPTArray, DPTBinary)
        assert isinstance(dpt_class.payload_length, int)
        assert dpt_class.payload_length, "Payload_length 0 is invalid"
        assert isinstance(dpt_class.dpt_main_number, int)
        assert dpt_class.dpt_main_number, "DPT main number 0 is invalid"
        assert isinstance(dpt_class.dpt_sub_number, int | type(None))
        assert isinstance(dpt_class.value_type, str)
        assert dpt_class.value_type, "Empty string for value_type is invalid"

    def test_validate_payload_array(self) -> None:
        """Test validate_payload method."""

        class DPTArrayTest(DPTBase):
            """Mock class for testing array payloads."""

            payload_type = DPTArray
            payload_length = 2

        with pytest.raises(CouldNotParseTelegram):
            DPTArrayTest.validate_payload(DPTArray((1,)))
        with pytest.raises(CouldNotParseTelegram):
            DPTArrayTest.validate_payload(DPTArray((1, 1, 1)))
        with pytest.raises(CouldNotParseTelegram):
            DPTArrayTest.validate_payload(DPTBinary(1))
        with pytest.raises(CouldNotParseTelegram):
            DPTArrayTest.validate_payload("why?")

        assert DPTArrayTest.validate_payload(DPTArray((1, 1))) == (1, 1)

    def test_validate_payload_binary(self) -> None:
        """Test validate_payload method."""

        class DPTBinaryTest(DPTBase):
            """Mock class for testing binary payloads."""

            payload_type = DPTBinary
            payload_length = 1

        with pytest.raises(CouldNotParseTelegram):
            DPTBinaryTest.validate_payload(DPTArray(1))
        with pytest.raises(CouldNotParseTelegram):
            DPTBinaryTest.validate_payload(DPTArray((1, 1)))
        with pytest.raises(CouldNotParseTelegram):
            DPTBinaryTest.validate_payload("why?")

        assert DPTBinaryTest.validate_payload(DPTBinary(1)) == (1,)


class TestDPTNumeric:
    """Test class for numeric transcoder base object."""

    @pytest.mark.parametrize("dpt_class", DPTNumeric.dpt_class_tree())
    def test_values(self, dpt_class: type[DPTNumeric]) -> None:
        """Test boundary values are set for numeric definitions (because mypy doesn't)."""

        assert isinstance(dpt_class.value_min, int | float)
        assert isinstance(dpt_class.value_max, int | float)
        assert isinstance(dpt_class.resolution, int | float)



"""KNX Constants used within io."""

from typing import Final

from xknx.telegram.address import IndividualAddress

DEFAULT_INDIVIDUAL_ADDRESS: Final = IndividualAddress("15.15.250")
DEFAULT_MCAST_GRP: Final = "224.0.23.12"
DEFAULT_MCAST_PORT: Final = 3671

CONNECTION_ALIVE_TIME: Final = 120
CONNECTIONSTATE_REQUEST_TIMEOUT: Final = 10
HEARTBEAT_RATE: Final = CONNECTION_ALIVE_TIME - (CONNECTIONSTATE_REQUEST_TIMEOUT * 5)

# Maximum time an authenticated secure session may remain unused (without
# any communication over this session) until the session will be dropped.
SESSION_TIMEOUT: Final = 60
SESSION_KEEPALIVE_RATE: Final = SESSION_TIMEOUT - 10
XKNX_SERIAL_NUMBER: Final = bytes.fromhex("00 00 78 6b 6e 78")



"""Implementation of KNX RGBW color data point type."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass

from .dpt import DPTComplex, DPTComplexData
from .payload import DPTArray, DPTBinary


@dataclass(slots=True)
class RGBWColor(DPTComplexData):
    """
    Representation of RGBW color.

    `red`: int 0..255; None if invalid
    `green`: int 0..255; None if invalid
    `blue`: int 0..255; None if invalid
    `white`: int 0..255; None if invalid
    """

    red: int | None = None
    green: int | None = None
    blue: int | None = None
    white: int | None = None

    @classmethod
    def from_dict(cls, data: Mapping[str, int]) -> RGBWColor:
        """Init from a dictionary."""
        result = {}
        for color in ("red", "green", "blue", "white"):
            try:
                value = data.get(color)
                result[color] = int(value) if value is not None else None
            except AttributeError as err:
                raise ValueError(f"Invalid value for color {color}: {err}") from err
            except ValueError as err:
                raise ValueError(f"Invalid value for color {color}: {err}") from err
        return cls(**result)

    def as_dict(self) -> dict[str, int | None]:
        """Create a JSON serializable dictionary."""
        return {
            "red": self.red,
            "green": self.green,
            "blue": self.blue,
            "white": self.white,
        }

    def __or__(self, other: RGBWColor) -> RGBWColor:
        """Merge two RGBWColor objects using only valid values."""
        return RGBWColor(
            red=other.red if other.red is not None else self.red,
            green=other.green if other.green is not None else self.green,
            blue=other.blue if other.blue is not None else self.blue,
            white=other.white if other.white is not None else self.white,
        )


class DPTColorRGBW(DPTComplex[RGBWColor]):
    """Abstraction for KNX 6 octet color RGBW (DPT 251.600)."""

    data_type = RGBWColor
    payload_type = DPTArray
    payload_length = 6
    dpt_main_number = 251
    dpt_sub_number = 600
    value_type = "color_rgbw"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> RGBWColor:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)
        red_valid = raw[5] >> 3 & 0b1
        green_valid = raw[5] >> 2 & 0b1
        blue_valid = raw[5] >> 1 & 0b1
        white_valid = raw[5] & 0b1
        return RGBWColor(
            red=raw[0] if red_valid else None,
            green=raw[1] if green_valid else None,
            blue=raw[2] if blue_valid else None,
            white=raw[3] if white_valid else None,
        )

    @classmethod
    def _to_knx(cls, value: RGBWColor) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        for color in (value.red, value.green, value.blue, value.white):
            if color is None:
                continue
            if not 0 <= color <= 255:
                raise ValueError(f"Color value out of range 0..255: {value}")
        return DPTArray(
            (
                value.red if value.red is not None else 0x00,
                value.green if value.green is not None else 0x00,
                value.blue if value.blue is not None else 0x00,
                value.white if value.white is not None else 0x00,
                0x00,  # reserved
                (
                    (value.red is not None) << 3
                    | (value.green is not None) << 2
                    | (value.blue is not None) << 1
                    | (value.white is not None)
                ),
            )
        )



"""Unit test for KNX/IP SecureWrapper objects."""

from xknx.knxip import KNXIPFrame, SecureWrapper


class TestKNXIPSecureWrapper:
    """Test class for KNX/IP SecureWrapper objects."""

    def test_secure_wrapper(self) -> None:
        """Test parsing and streaming secure wrapper KNX/IP packet."""
        sequence_number = bytes.fromhex("00 00 00 00 00 00")
        knx_serial_number = bytes.fromhex("00 fa 12 34 56 78")
        message_tag = bytes.fromhex("af fe")
        encrypted_data = bytes.fromhex(
            "79 15 a4 f3 6e 6e 42 08"  # SessionAuthenticate Frame
            "d2 8b 4a 20 7d 8f 35 c0"
            "d1 38 c2 6a 7b 5e 71 69"
        )
        message_authentication_code = bytes.fromhex(
            "52 db a8 e7 e4 bd 80 bd 7d 86 8a 3a e7 87 49 de"
        )
        raw = (
            bytes.fromhex(
                "06 10 09 50 00 3e"  # KNXnet/IP header
                "00 01"  # Secure Session Identifier
            )
            + sequence_number
            + knx_serial_number
            + message_tag
            + encrypted_data
            + message_authentication_code
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SecureWrapper)
        assert knxipframe.body.secure_session_id == 1
        assert knxipframe.body.sequence_information == sequence_number
        assert knxipframe.body.serial_number == knx_serial_number
        assert knxipframe.body.message_tag == message_tag
        assert knxipframe.body.encrypted_data == encrypted_data
        assert (
            knxipframe.body.message_authentication_code == message_authentication_code
        )

        assert knxipframe.to_knx() == raw

        secure_wrapper = SecureWrapper(
            secure_session_id=1,
            sequence_information=sequence_number,
            serial_number=knx_serial_number,
            message_tag=message_tag,
            encrypted_data=encrypted_data,
            message_authentication_code=message_authentication_code,
        )
        knxipframe2 = KNXIPFrame.init_from_body(secure_wrapper)
        assert knxipframe2.to_knx() == raw



"""
Module for serialization and deserialization of KNX DIB information.

DIB is Description Information Block.

A KNX/IP Search Response may contain several DIBs of different types:

* DIBSuppSVCFamilies:   Supported features of device
* DIBDeviceInformation: Name, serial number, some unimportant flags
* DIBGeneric:           General Information
                        (fallback for unknown dib type codes)
"""

from __future__ import annotations

from abc import ABC, abstractmethod
import socket
from typing import NamedTuple, final

from xknx.exceptions import CouldNotParseKNXIP
from xknx.telegram import IndividualAddress

from .knxip_enum import DIBServiceFamily, DIBTypeCode, KNXMedium

DIB_HEADER_LENGTH = 2  # structure length and description type code


class DIB(ABC):
    """
    Base class for DIB (Description Information Block).

    This base class is only the interface for the derived
    classes.
    """

    @abstractmethod
    def calculated_length(self) -> int:
        """Get length of KNX/IP object."""
        # The structure shall always have an even number of octets which may have to be
        # achieved by padding with 00h in the last octet of the DIB structure.

    @abstractmethod
    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""

    @abstractmethod
    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""

    @staticmethod
    def determine_dib(raw: bytes) -> DIB:
        """Determine dib type out of dib type code."""
        if len(raw) < 2:
            raise CouldNotParseKNXIP("could not parse DIB header")
        dtc = DIBTypeCode(raw[1])

        if dtc == DIBTypeCode.DEVICE_INFO:
            return DIBDeviceInformation()
        if dtc == DIBTypeCode.SUPP_SVC_FAMILIES:
            return DIBSuppSVCFamilies()
        if dtc == DIBTypeCode.SECURED_SERVICE_FAMILIES:
            return DIBSecuredServiceFamilies()
        if dtc == DIBTypeCode.TUNNELING_INFO:
            return DIBTunnelingInfo()
        return DIBGeneric()


class DIBGeneric(DIB):
    """
    Module for serialization and deserialization of KNX DIB Generic.

    Fallback for not implemented DIBTypeCodes.
    """

    def __init__(self) -> None:
        """Initialize DIBGeneric class."""
        # DTC Description Type Code
        self.dtc: DIBTypeCode | int = 0
        # IBD Information Block Data
        self.data = b""

    def calculated_length(self) -> int:
        """Get length of KNX/IP object."""
        data_length = len(self.data)
        return DIB_HEADER_LENGTH + data_length + data_length % 2

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < 2:
            raise CouldNotParseKNXIP("could not parse DIB header")

        dib_length = raw[0]
        if len(raw) < dib_length:
            raise CouldNotParseKNXIP("DIB wrong length")
        try:
            self.dtc = DIBTypeCode(raw[1])
        except ValueError:
            self.dtc = raw[1]
        self.data = raw[2:dib_length]

        return dib_length

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        if not isinstance(self.dtc, DIBTypeCode):
            try:
                self.dtc = DIBTypeCode(self.dtc)
            except ValueError:
                raise CouldNotParseKNXIP("DTC invalid") from None
        return (
            bytes((self.calculated_length(), self.dtc.value))
            + self.data
            + bytes(len(self.data) % 2)  # padding
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return f'<DIB dtc="{self.dtc}" data="{", ".join(f"0x{i:02x}" for i in self.data)}" />'


@final
class DIBDeviceInformation(DIB):
    """Class for serialization and deserialization of KNX DIB Device Information Block."""

    LENGTH = 54

    def __init__(self) -> None:
        """Initialize DIBDeviceInformation class."""
        self.knx_medium: KNXMedium = KNXMedium.TP1
        self.programming_mode: bool = False
        self.individual_address: IndividualAddress = IndividualAddress(0)
        self.installation_number: int = 0
        self.project_number: int = 0
        self.serial_number: str = ""
        self.multicast_address: str = "224.0.23.12"
        self.mac_address: str = ""
        self.name: str = ""

    def calculated_length(self) -> int:
        """Get length of KNX/IP object."""
        return DIBDeviceInformation.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < DIBDeviceInformation.LENGTH:
            raise CouldNotParseKNXIP("wrong connection header length")
        if raw[0] != DIBDeviceInformation.LENGTH:
            raise CouldNotParseKNXIP("wrong connection header length")
        if DIBTypeCode(raw[1]) != DIBTypeCode.DEVICE_INFO:
            raise CouldNotParseKNXIP("DIB is no device info")

        self.knx_medium = KNXMedium(raw[2])
        # last bit of device_status. All other bits are unused
        self.programming_mode = bool(raw[3])
        self.individual_address = IndividualAddress.from_knx(raw[4:6])
        installation_project_identifier = raw[6] * 256 + raw[7]
        self.project_number = installation_project_identifier >> 4
        self.installation_number = installation_project_identifier & 15
        self.serial_number = raw[8:14].hex(":")
        self.multicast_address = socket.inet_ntoa(raw[14:18])
        self.mac_address = raw[18:24].hex(":")
        self.name = raw[24:54].decode(encoding="latin_1", errors="replace").rstrip("\0")
        return DIBDeviceInformation.LENGTH

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""

        def hex_notation_to_knx(colon_hex: str) -> bytes:
            """Serialize hex notation."""
            return bytes.fromhex(colon_hex.replace(":", ""))

        def ip_to_knx(ip_addr: str) -> bytes:
            """Serialize ip."""
            return socket.inet_aton(ip_addr)

        def name_str_to_knx(string: str) -> bytes:
            """Serialize name string."""
            # pad with null bytes to length 30; ISO 8859-1 (latin_1) according to KNX specification
            return bytes(string[:30], "latin_1").ljust(30, b"\0")

        installation_project_identifier = (
            (self.project_number * 16) + self.installation_number
        ).to_bytes(2, "big")

        return (
            bytes(
                (
                    DIBDeviceInformation.LENGTH,
                    DIBTypeCode.DEVICE_INFO.value,
                    self.knx_medium.value,
                    self.programming_mode,
                )
            )
            + self.individual_address.to_knx()
            + installation_project_identifier
            + hex_notation_to_knx(self.serial_number)
            + ip_to_knx(self.multicast_address)
            + hex_notation_to_knx(self.mac_address)
            + name_str_to_knx(self.name)
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<DIBDeviceInformation "
            f'\n\tknx_medium="{self.knx_medium}" '
            f'\n\tprogramming_mode="{self.programming_mode}" '
            f'\n\tindividual_address="{self.individual_address}" '
            f'\n\tinstallation_number="{self.installation_number}" '
            f'\n\tproject_number="{self.project_number}" '
            f'\n\tserial_number="{self.serial_number}" '
            f'\n\tmulticast_address="{self.multicast_address}" '
            f'\n\tmac_address="{self.mac_address}" '
            f'\n\tname="{self.name}" />'
        )


class _DIBServiceFamilies(DIB):
    """Base class for serialization and deserialization of KNX DIB Service Families."""

    type_code: DIBTypeCode

    class Family:
        """Class for storing a supported device family."""

        def __init__(self, name: DIBServiceFamily, version: int) -> None:
            """Initialize DIBSuppSVCFamilies.Family."""
            self.name = name
            self.version = version

        def to_knx(self) -> bytes:
            """Serialize to KNX/IP raw data."""
            return bytes((self.name.value, self.version))

        def __repr__(self) -> str:
            """Return object as readable string."""
            return f'<Family name="{self.name}" version="{self.version}" />'

        def __eq__(self, other: object) -> bool:
            """Equal operator."""
            return self.__dict__ == other.__dict__

    def __init__(self) -> None:
        """Initialize DIBSuppSVCFamilies class."""
        self.families: list[DIBSuppSVCFamilies.Family] = []

    def supports(self, name: DIBServiceFamily, version: int | None = None) -> bool:
        """Return if device supports a given service family by name and optional minimum version."""
        return any(
            name == family.name and (version is None or family.version >= version)
            for family in self.families
        )

    def version(self, name: DIBServiceFamily) -> int | None:
        """Return version of a given service family."""
        return next(
            (family.version for family in self.families if name == family.name),
            None,
        )

    def calculated_length(self) -> int:
        """Get length of KNX/IP object."""
        return len(self.families) * 2 + DIB_HEADER_LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < 2:
            raise CouldNotParseKNXIP("DIB header too small")
        length = raw[0]
        if (len(raw) < length) or (length % 2):
            raise CouldNotParseKNXIP("DIB wrong size")
        if DIBTypeCode(raw[1]) != self.type_code:
            raise CouldNotParseKNXIP(
                f"DIB has wrong type code for {self.__class__.__name__}"
            )

        for pos in range(2, length, 2):
            name = DIBServiceFamily(raw[pos])
            version = raw[pos + 1]
            self.families.append(DIBSuppSVCFamilies.Family(name, version))
        return length

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return bytes(
            (
                self.calculated_length(),
                self.type_code.value,
            )
        ) + b"".join(family.to_knx() for family in self.families)

    def __repr__(self) -> str:
        """Return object as readable string."""
        _families_str = ", ".join(
            f"{family.name} version: {family.version}" for family in self.families
        )
        return f'<{self.__class__.__name__} families="[{_families_str}]" />'


@final
class DIBSuppSVCFamilies(_DIBServiceFamilies):
    """Class for serialization and deserialization of KNX DIB Supported Services."""

    type_code = DIBTypeCode.SUPP_SVC_FAMILIES


@final
class DIBSecuredServiceFamilies(_DIBServiceFamilies):
    """Class for serialization and deserialization of KNX DIB Secured Service Families."""

    type_code = DIBTypeCode.SECURED_SERVICE_FAMILIES


class TunnelingSlotStatus(NamedTuple):
    """Class for storing tunneling slot status."""

    usable: bool
    authorized: bool
    free: bool

    def __bytes__(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return bytes(
            (
                0x00,  # reserved
                self.usable << 2 | self.authorized << 1 | self.free,
            )
        )


@final
class DIBTunnelingInfo(DIB):
    """Class for serialization and deserialization of KNX DIB Tunneling Info."""

    def __init__(
        self, slots: dict[IndividualAddress, TunnelingSlotStatus] | None = None
    ) -> None:
        """Initialize DIBTunnelingInfo class."""
        self.max_apdu_length = 248
        self.slots = slots or {}

    def calculated_length(self) -> int:
        """Get length of KNX/IP object."""
        return 2 + 2 + len(self.slots) * 4

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < 4:
            raise CouldNotParseKNXIP("DIB header too small")
        length = raw[0]
        if (len(raw) < length) or (length % 4):
            raise CouldNotParseKNXIP("DIB wrong size")
        if DIBTypeCode(raw[1]) != DIBTypeCode.TUNNELING_INFO:
            raise CouldNotParseKNXIP(
                f"DIB has wrong type code for {self.__class__.__name__}"
            )

        self.max_apdu_length = int.from_bytes(raw[2:4], "big")
        for pos in range(4, length, 4):
            address = IndividualAddress.from_knx(raw[pos : pos + 2])
            status = TunnelingSlotStatus(
                usable=bool(raw[pos + 3] >> 2 & 0b1),
                authorized=bool(raw[pos + 3] >> 1 & 0b1),
                free=bool(raw[pos + 3] & 0b1),
            )
            self.slots[address] = status
        return length

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            bytes((self.calculated_length(), DIBTypeCode.TUNNELING_INFO.value))
            + self.max_apdu_length.to_bytes(2, "big")
            + b"".join(
                address.to_knx() + bytes(status)
                for address, status in self.slots.items()
            )
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            f"<{self.__class__.__name__} max_adpu_lenght={self.max_apdu_length} "
            f"slots={self.slots}/>"
        )



"""
Module for managing a remote value typically used within a sensor.

The module maps a given value_type to a DPT class and uses this class
for serialization and deserialization of the KNX value.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from xknx.dpt import DPTBase, DPTNumeric, DPTString
from xknx.exceptions import ConversionError
from xknx.typing import DPTParsable

from .remote_value import GroupAddressesType, RemoteValue, RVCallbackType

if TYPE_CHECKING:
    from xknx.xknx import XKNX

ValueT = TypeVar("ValueT")


class _RemoteValueGeneric(RemoteValue[ValueT]):
    """Abstraction for generic DPT types."""

    dpt_base_class: type[DPTBase]
    _default_dpt_class: type[DPTBase] | None = None
    dpt_class: type[DPTBase]

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        value_type: DPTParsable | type[DPTBase] | None = None,
        device_name: str | None = None,
        feature_name: str = "Value",
        after_update_cb: RVCallbackType[ValueT] | None = None,
    ) -> None:
        """Initialize RemoteValueSensor class."""
        try:
            if value_type is None:
                if self._default_dpt_class is None:
                    raise ValueError
                self.dpt_class = self._default_dpt_class
            else:
                self.dpt_class = self.dpt_base_class.get_dpt(value_type)
        except ValueError:
            raise ConversionError(
                f"invalid value type for base class {self.dpt_base_class.dpt_name()}",
                value_type=value_type,
                device_name=device_name,
            ) from None
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    @property
    def unit_of_measurement(self) -> str | None:
        """Return the unit of measurement."""
        return self.dpt_class.unit

    @property
    def ha_device_class(self) -> str | None:
        """Return a string representing the home assistant device class."""
        return getattr(self.dpt_class, "ha_device_class", None)


class RemoteValueSensor(_RemoteValueGeneric[int | float | str]):
    """Abstraction for sensor DPT types."""

    dpt_base_class = DPTBase
    dpt_class: type[DPTBase]


class RemoteValueNumeric(_RemoteValueGeneric[int | float]):
    """Abstraction for numeric DPT types."""

    dpt_base_class = DPTNumeric
    dpt_class: type[DPTNumeric]


class RemoteValueString(_RemoteValueGeneric[str]):
    """Abstraction for string DPT types."""

    dpt_base_class = DPTString
    dpt_class: type[DPTString]
    _default_dpt_class = DPTString



"""
Module for Serialization and Deserialization of a KNX Connect Response information.

Connect requests are used to start a new tunnel connection on a KNX/IP device.
With a Connect Response the receiving party acknowledges the valid processing of the request,
assigns a communication channel and an individual address for the client.
"""

from __future__ import annotations

from xknx.exceptions import CouldNotParseKNXIP
from xknx.telegram import IndividualAddress

from .body import KNXIPBodyResponse
from .error_code import ErrorCode
from .hpai import HPAI
from .knxip_enum import ConnectRequestType, KNXIPServiceType


class ConnectResponse(KNXIPBodyResponse):
    """Representation of a KNX Connect Response."""

    SERVICE_TYPE = KNXIPServiceType.CONNECT_RESPONSE

    def __init__(
        self,
        communication_channel: int = 0,
        status_code: ErrorCode = ErrorCode.E_NO_ERROR,
        data_endpoint: HPAI | None = None,
        crd: ConnectResponseData | None = None,
    ) -> None:
        """Initialize ConnectResponse class."""
        self.communication_channel = communication_channel
        self.status_code = status_code
        self.data_endpoint = data_endpoint or HPAI()
        self.crd = crd or ConnectResponseData()

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return 2 + HPAI.LENGTH + self.crd.calculated_length()

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""

        self.communication_channel = raw[0]
        self.status_code = ErrorCode(raw[1])
        pos = 2

        if self.status_code == ErrorCode.E_NO_ERROR:
            pos += self.data_endpoint.from_knx(raw[pos:])
            pos += self.crd.from_knx(raw[pos:])
        else:
            # do not parse HPAI and CRD in case of errors - just check length
            pos = len(raw)
        return pos

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""

        return (
            bytes((self.communication_channel, self.status_code.value))
            + self.data_endpoint.to_knx()
            + self.crd.to_knx()
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<ConnectResponse "
            f'communication_channel="{self.communication_channel}" '
            f'status_code="{self.status_code}" '
            f'data_endpoint="{self.data_endpoint}" '
            f'crd="{self.crd}" />'
        )


class ConnectResponseData:
    """Representation of a KNX Connect Response Data block (CRD)."""

    CRD_LENGTH = 2
    CRD_TUNNEL_LENGTH = 4

    def __init__(
        self,
        request_type: ConnectRequestType = ConnectRequestType.TUNNEL_CONNECTION,
        individual_address: IndividualAddress | None = None,
    ) -> None:
        """Initialize ConnectResponseData object."""
        self.request_type = request_type
        self.individual_address = individual_address

    def _is_tunnel_crd(self) -> bool:
        return self.request_type == ConnectRequestType.TUNNEL_CONNECTION

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return (
            ConnectResponseData.CRD_TUNNEL_LENGTH
            if self._is_tunnel_crd()
            else ConnectResponseData.CRD_LENGTH
        )

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        crd_length = raw[0]
        if len(raw) < crd_length:
            raise CouldNotParseKNXIP("CRD has wrong length")
        if crd_length < ConnectResponseData.CRD_LENGTH:
            raise CouldNotParseKNXIP("CRD length too small")
        self.request_type = ConnectRequestType(raw[1])
        if self._is_tunnel_crd():
            if crd_length != ConnectResponseData.CRD_TUNNEL_LENGTH:
                raise CouldNotParseKNXIP("CRD has wrong length")
            self.individual_address = IndividualAddress.from_knx(raw[2:4])
        elif crd_length != ConnectResponseData.CRD_LENGTH:
            raise CouldNotParseKNXIP("CRD has wrong length")
        return crd_length

    def to_knx(self) -> bytes:
        """Serialize CRD (Connect Response Data Block)."""
        _crd = bytes(
            (
                self.calculated_length(),
                self.request_type.value,
            )
        )
        if self._is_tunnel_crd():
            assert self.individual_address is not None
            return _crd + self.individual_address.to_knx()

        return _crd

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        return self.__dict__ == other.__dict__

    def __repr__(self) -> str:
        """Return object as readable string."""
        _address = (
            f'individual_address="{self.individual_address}" '
            if self._is_tunnel_crd() and self.individual_address
            else ""
        )
        return f'<ConnectResponseData request_type="{self.request_type}" {_address}/>'



"""Example for Climate device."""

import asyncio

from xknx import XKNX
from xknx.devices import Climate


async def main() -> None:
    """Connect to KNX/IP and read the state of a Climate device."""
    xknx = XKNX()
    async with xknx:
        climate = Climate(xknx, "TestClimate", group_address_temperature="6/2/1")
        xknx.devices.async_add(climate)
        await climate.sync(wait_for_result=True)
        # Will print out state of climate including current temperature:
        print(climate)


asyncio.run(main())



"""Unit test for KNX/IP SessionStatus objects."""

from xknx.knxip import KNXIPFrame, SessionStatus
from xknx.knxip.knxip_enum import SecureSessionStatusCode


class TestKNXIPSessionStatus:
    """Test class for KNX/IP SessionStatus objects."""

    def test_session_status(self) -> None:
        """Test parsing and streaming session status KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 09 54 00 08"  # KNXnet/IP header
            "00"  # status code 00h STATUS_AUTHENTICATION_SUCCESS
            "00"  # reserved
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SessionStatus)
        assert (
            knxipframe.body.status
            == SecureSessionStatusCode.STATUS_AUTHENTICATION_SUCCESS
        )

        assert knxipframe.to_knx() == raw

        session_status = SessionStatus(
            status=SecureSessionStatusCode.STATUS_AUTHENTICATION_SUCCESS
        )
        knxipframe2 = KNXIPFrame.init_from_body(session_status)

        assert knxipframe2.to_knx() == raw



"""Enums for KNX Application Layer."""

from enum import Enum


class ReturnCode(Enum):
    """Enum class for Generic device management Return Codes."""

    ## Basic positive Return Code
    # The service, function or command is executed successfully, without additional information.
    E_SUCCESS = 0x00
    ## Generic negative Return Codes
    # Memory cannot be accessed or only with fault(s).
    E_MEMORY_ERROR = 0xF1
    # Requested data will not fit into a Frame supported by this server.
    # This shall be used for Device limitations of the maximum supported Frame length
    # by accessing resources (Properties, Function Properties, memory…) of the device.
    E_LENGTH_EXCEEDS_MAX_APDU_LENGTH = 0xF4
    # Writing data beyond what is reserved for the addressed Resource.
    E_DATA_OVERFLOW = 0xF5
    # Write value too low. Preferable to give this instead of “Value not supported”.
    E_DATA_MIN = 0xF6
    # Write value too high. Preferable to give this instead of “Value not supported”.
    E_DATA_MAX = 0xF7
    # The service or function is supported, but request data is not valid for this receiver.
    E_DATA_VOID = 0xF8
    # Data could generally be written, but not possible at this time.
    E_TEMPORARILY_NOT_AVAILABLE = 0xF9
    # Read access attempted to a “write only” service or Resource.
    E_ACCESS_WRITE_ONLY = 0xFA
    # Write access attempted to a “read only” service or Resource.
    E_ACCESS_READ_ONLY = 0xFB
    # Access denied due to authorization reasons. A_Authorize as well as KNX Security
    E_ACCESS_DENIED = 0xFC
    # Interface Object or Property is not present, or index is out of range.
    E_ADDRESS_VOID = 0xFD
    # Write access with a wrong datatype (Datapoint length).
    E_DATA_TYPE_CONFLICT = 0xFE
    # The service, function or command has failed without a closer indication of the problem.
    E_ERROR = 0xFF
    ## Generic positive Return Codes
    # (01h-1Fh - None proposed)
    ## Specific positive Return Codes
    # (20h-5Fh - None proposed)
    ## Specific negative Return Codes
    # (A0h-DFh - None proposed)



"""Unit test for RemoteValueSetpointShift objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary, DPTTemperature, DPTValue1Count
from xknx.exceptions import ConversionError, CouldNotParseTelegram
from xknx.remote_value.remote_value_setpoint_shift import (
    RemoteValueSetpointShift,
    SetpointShiftMode,
)


class TestRemoteValueSetpointShift:
    """Test class for RemoteValueSetpointShift objects."""

    def test_payload_valid_mode_assignment(self) -> None:
        """Test if setpoint_shift_mode is assigned properly by payload length."""
        xknx = XKNX()
        remote_value = RemoteValueSetpointShift(xknx=xknx)
        dpt_6_payload = DPTValue1Count.to_knx(1)
        dpt_9_payload = DPTTemperature.to_knx(1)

        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(DPTBinary(0))
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(DPTArray((0, 1, 2)))

        # DPT 6 - payload_length 1
        assert remote_value._internal_dpt_class is None
        # default setpoint_shift_step = 0.1
        assert remote_value.from_knx(dpt_6_payload) == 0.1
        assert remote_value._internal_dpt_class == SetpointShiftMode.DPT6010.value
        with pytest.raises(CouldNotParseTelegram):
            # DPT 9 is invalid now
            remote_value.from_knx(dpt_9_payload)

        remote_value._internal_dpt_class = None
        # DPT 9 - payload_length 2
        assert remote_value.from_knx(dpt_9_payload) == 1
        assert remote_value._internal_dpt_class == SetpointShiftMode.DPT9002.value
        with pytest.raises(CouldNotParseTelegram):
            # DPT 6 is invalid now
            remote_value.from_knx(dpt_6_payload)

    def test_payload_valid_preassigned_mode(self) -> None:
        """Test if setpoint_shift_mode is assigned properly by payload length."""
        xknx = XKNX()
        remote_value_6 = RemoteValueSetpointShift(
            xknx=xknx, setpoint_shift_mode=SetpointShiftMode.DPT6010
        )
        remote_value_9 = RemoteValueSetpointShift(
            xknx=xknx, setpoint_shift_mode=SetpointShiftMode.DPT9002
        )
        dpt_6_payload = DPTValue1Count.to_knx(1)
        dpt_9_payload = DPTTemperature.to_knx(1)

        assert remote_value_6._internal_dpt_class == DPTValue1Count
        with pytest.raises(CouldNotParseTelegram):
            remote_value_6.from_knx(None)
        with pytest.raises(CouldNotParseTelegram):
            remote_value_6.from_knx(dpt_9_payload)
        with pytest.raises(CouldNotParseTelegram):
            remote_value_6.from_knx(DPTArray((1, 2, 3, 4)))
        with pytest.raises(CouldNotParseTelegram):
            remote_value_6.from_knx(DPTBinary(1))
        assert remote_value_6.from_knx(dpt_6_payload) == 0.1

        assert remote_value_9._internal_dpt_class == DPTTemperature
        with pytest.raises(CouldNotParseTelegram):
            remote_value_9.from_knx(None)
        with pytest.raises(CouldNotParseTelegram):
            remote_value_9.from_knx(dpt_6_payload)
        with pytest.raises(CouldNotParseTelegram):
            remote_value_9.from_knx(DPTArray((1, 2, 3)))
        with pytest.raises(CouldNotParseTelegram):
            remote_value_9.from_knx(DPTBinary(1))
        assert remote_value_9.from_knx(dpt_9_payload) == 1

    def test_to_knx_uninitialized(self) -> None:
        """Test to_knx raising ConversionError."""
        xknx = XKNX()
        remote_value = RemoteValueSetpointShift(xknx=xknx)

        assert remote_value._internal_dpt_class is None
        with pytest.raises(ConversionError):
            remote_value.to_knx(1)

    def test_to_knx_dpt_6(self) -> None:
        """Test to_knx returning DPT 6.010 payload."""
        xknx = XKNX()
        remote_value = RemoteValueSetpointShift(
            xknx=xknx, setpoint_shift_mode=SetpointShiftMode.DPT6010
        )
        assert remote_value.setpoint_shift_step == 0.1
        assert remote_value.to_knx(1) == DPTArray((10,))

    def test_to_knx_dpt_9(self) -> None:
        """Test to_knx returning DPT 9.002 payload."""
        xknx = XKNX()
        remote_value = RemoteValueSetpointShift(
            xknx=xknx, setpoint_shift_mode=SetpointShiftMode.DPT9002
        )
        assert remote_value.to_knx(1) == DPTArray((0x00, 0x64))

    def test_from_knx_uninitialized(self) -> None:
        """Test from_knx for uninitialized setpoint_shift_mode."""
        xknx = XKNX()
        remote_value = RemoteValueSetpointShift(xknx=xknx)
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(1)
        # assign DPT 9.002 mode
        assert remote_value.from_knx(DPTArray((0x00, 0x64))) == 1
        assert remote_value.from_knx(DPTArray((0x07, 0xD0))) == 20
        # wrong payload length raises, once assigned
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(DPTArray((10,)))

    def test_from_knx_dpt_6(self) -> None:
        """Test from_knx for DPT 6.010 setpoint_shift_mode."""
        xknx = XKNX()
        remote_value = RemoteValueSetpointShift(
            xknx=xknx, setpoint_shift_mode=SetpointShiftMode.DPT6010
        )
        assert remote_value.setpoint_shift_step == 0.1
        assert remote_value.from_knx(DPTArray((10,))) == 1

    def test_from_knx_dpt_9(self) -> None:
        """Test from_knx for DPT 9.002 setpoint_shift_mode."""
        xknx = XKNX()
        remote_value = RemoteValueSetpointShift(
            xknx=xknx, setpoint_shift_mode=SetpointShiftMode.DPT9002
        )
        assert remote_value.from_knx(DPTArray((0x00, 0x64))) == 1



"""
Module for managing a sensor via KNX.

It provides functionality for

* reading the current state from KNX bus.
* watching for state updates from KNX bus.
"""

from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Any

from xknx.dpt import DPTBase
from xknx.remote_value import (
    GroupAddressesType,
    RemoteValue,
    RemoteValueSensor,
)
from xknx.typing import DPTParsable

from .device import Device, DeviceCallbackType

if TYPE_CHECKING:
    from xknx.telegram import Telegram
    from xknx.xknx import XKNX


class Sensor(Device):
    """Class for managing a sensor."""

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        always_callback: bool = False,
        value_type: DPTParsable | type[DPTBase] | None = None,
        device_updated_cb: DeviceCallbackType[Sensor] | None = None,
    ) -> None:
        """Initialize Sensor class."""
        super().__init__(xknx, name, device_updated_cb)
        self.sensor_value = RemoteValueSensor(
            xknx,
            group_address_state=group_address_state,
            sync_state=sync_state,
            value_type=value_type,
            device_name=self.name,
            after_update_cb=self.after_update,
        )
        self.always_callback = always_callback

    def _iter_remote_values(self) -> Iterator[RemoteValue[Any]]:
        """Iterate the devices RemoteValue classes."""
        yield self.sensor_value

    @property
    def last_telegram(self) -> Telegram | None:
        """Return the last telegram received from the RemoteValue."""
        return self.sensor_value.telegram

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming and outgoing GROUP WRITE telegram."""
        self.sensor_value.process(telegram, always_callback=self.always_callback)

    def process_group_response(self, telegram: Telegram) -> None:
        """Process incoming GroupValueResponse telegrams."""
        self.sensor_value.process(telegram, always_callback=self.always_callback)

    def unit_of_measurement(self) -> str | None:
        """Return the unit of measurement."""
        return self.sensor_value.unit_of_measurement

    def ha_device_class(self) -> str | None:
        """Return the home assistant device class as string."""
        return self.sensor_value.ha_device_class

    def resolve_state(self) -> Any | None:
        """Return the current state of the sensor as a human readable string."""
        return self.sensor_value.value

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f'<Sensor name="{self.name}" '
            f"sensor={self.sensor_value.group_addr_str()} "
            f"value={self.resolve_state().__repr__()} "
            f'unit="{self.unit_of_measurement()}"/>'
        )



"""Unit test for KNX DPT 5.010 value."""

import pytest

from xknx.dpt import DPTAngle, DPTArray, DPTScaling, DPTTariff, DPTValue1Ucount
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPTValue1Ucount:
    """Test class for KNX 8-bit unsigned value."""

    def test_value_50(self) -> None:
        """Test parsing and streaming of DPTValue1Ucount 50."""
        assert DPTValue1Ucount.to_knx(50) == DPTArray(0x32)
        assert DPTValue1Ucount.from_knx(DPTArray((0x32,))) == 50

    def test_value_max(self) -> None:
        """Test parsing and streaming of DPTValue1Ucount 255."""
        assert DPTValue1Ucount.to_knx(255) == DPTArray(0xFF)
        assert DPTValue1Ucount.from_knx(DPTArray((0xFF,))) == 255

    def test_value_min(self) -> None:
        """Test parsing and streaming of DPTValue1Ucount 0."""
        assert DPTValue1Ucount.to_knx(0) == DPTArray(0x00)
        assert DPTValue1Ucount.from_knx(DPTArray((0x00,))) == 0

    def test_to_knx_min_exceeded(self) -> None:
        """Test parsing of DPTValue1Ucount with wrong value (underflow)."""
        with pytest.raises(ConversionError):
            DPTValue1Ucount.to_knx(DPTValue1Ucount.value_min - 1)

    def test_to_knx_max_exceeded(self) -> None:
        """Test parsing of DPTValue1Ucount with wrong value (overflow)."""
        with pytest.raises(ConversionError):
            DPTValue1Ucount.to_knx(DPTValue1Ucount.value_max + 1)

    def test_to_knx_wrong_parameter(self) -> None:
        """Test parsing of DPTValue1Ucount with wrong value (string)."""
        with pytest.raises(ConversionError):
            DPTValue1Ucount.to_knx("fnord")

    def test_from_knx_wrong_parameter(self) -> None:
        """Test parsing of DPTValue1Ucount with wrong value (3 byte array)."""
        with pytest.raises(CouldNotParseTelegram):
            DPTValue1Ucount.from_knx(DPTArray((0x01, 0x02, 0x03)))

    def test_from_knx_wrong_value(self) -> None:
        """Test parsing of DPTValue1Ucount with value which exceeds limits."""
        with pytest.raises(ConversionError):
            DPTValue1Ucount.from_knx(DPTArray((0x256,)))

    def test_from_knx_wrong_parameter2(self) -> None:
        """Test parsing of DPTValue1Ucount with wrong value (array containing string)."""
        with pytest.raises(CouldNotParseTelegram):
            DPTValue1Ucount.from_knx("0x23")


class TestDPTScaling:
    """Test class for KNX scaling value."""

    @pytest.mark.parametrize(
        ("raw", "value"),
        [
            ((0x4C,), 30),
            ((0xFC,), 99),
            ((0xFF,), 100),
            ((0x00,), 0),
        ],
    )
    def test_transcoder(self, raw: tuple[int], value: int) -> None:
        """Test parsing and streaming of DPTScaling."""
        assert DPTScaling.to_knx(value) == DPTArray(raw)
        assert DPTScaling.from_knx(DPTArray(raw)) == value

    def test_to_knx_min_exceeded(self) -> None:
        """Test parsing of DPTScaling with wrong value (underflow)."""
        with pytest.raises(ConversionError):
            DPTScaling.to_knx(-1)

    def test_to_knx_max_exceeded(self) -> None:
        """Test parsing of DPTScaling with wrong value (overflow)."""
        with pytest.raises(ConversionError):
            DPTScaling.to_knx(101)

    def test_to_knx_wrong_parameter(self) -> None:
        """Test parsing of DPTScaling with wrong value (string)."""
        with pytest.raises(ConversionError):
            DPTScaling.to_knx("fnord")

    def test_from_knx_wrong_parameter(self) -> None:
        """Test parsing of DPTScaling with wrong value (3 byte array)."""
        with pytest.raises(CouldNotParseTelegram):
            DPTScaling.from_knx(DPTArray((0x01, 0x02, 0x03)))

    def test_from_knx_wrong_value(self) -> None:
        """Test parsing of DPTScaling with value which exceeds limits."""
        with pytest.raises(ConversionError):
            DPTScaling.from_knx(DPTArray((0x256,)))


class TestDPTAngle:
    """Test class for KNX scaling value."""

    @pytest.mark.parametrize(
        ("raw", "value"),
        [
            ((0x15,), 30),
            ((0xBF,), 270),
            ((0xFF,), 360),
            ((0x00,), 0),
        ],
    )
    def test_transcoder(self, raw: tuple[int], value: int) -> None:
        """Test parsing and streaming of DPTAngle."""
        assert DPTAngle.to_knx(value) == DPTArray(raw)
        assert DPTAngle.from_knx(DPTArray(raw)) == value

    def test_to_knx_min_exceeded(self) -> None:
        """Test parsing of DPTAngle with wrong value (underflow)."""
        with pytest.raises(ConversionError):
            DPTAngle.to_knx(-1)

    def test_to_knx_max_exceeded(self) -> None:
        """Test parsing of DPTAngle with wrong value (overflow)."""
        with pytest.raises(ConversionError):
            DPTAngle.to_knx(361)


class TestDPTTariff:
    """Test class for KNX 8-bit tariff information."""

    def test_from_knx_max_exceeded(self) -> None:
        """Test parsing of DPTTariff with wrong value (overflow)."""
        with pytest.raises(ConversionError):
            DPTTariff.from_knx(DPTArray((0xFF,)))



"""Unit test for KNX/IP Disconnect Request/Response."""

import asyncio
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from xknx.io.request_response import RequestResponse
from xknx.io.transport import UDPTransport
from xknx.knxip import DisconnectResponse, KNXIPBody


class TestConnectResponse:
    """Test class for xknx/io/Disconnect objects."""

    async def test_create_knxipframe_err(self) -> None:
        """Test if create_knxipframe of base class raises an exception."""
        udp_transport = UDPTransport(("192.168.1.1", 0), ("192.168.1.2", 1234))
        request_response = RequestResponse(udp_transport, DisconnectResponse)
        request_response.timeout_in_seconds = 0

        with pytest.raises(NotImplementedError):
            await request_response.start()

    @patch("logging.Logger.debug")
    @patch(
        "xknx.io.request_response.RequestResponse.send_request", new_callable=AsyncMock
    )
    async def test_request_response_timeout(
        self, _send_request_mock: MagicMock, logger_debug_mock: AsyncMock
    ) -> None:
        """Test RequestResponse: timeout. No callback shall be left."""
        udp_transport = UDPTransport(("192.168.1.1", 0), ("192.168.1.2", 1234))
        requ_resp = RequestResponse(udp_transport, KNXIPBody)
        requ_resp.response_received_event.wait = MagicMock(
            side_effect=asyncio.TimeoutError()
        )
        await requ_resp.start()
        # Debug message was logged
        logger_debug_mock.assert_called_once_with(
            "Error: KNX bus did not respond in time (%s secs) to request of type '%s'",
            1.0,
            "RequestResponse",
        )
        # Callback was removed again
        assert not udp_transport.callbacks

    @patch(
        "xknx.io.request_response.RequestResponse.send_request", new_callable=AsyncMock
    )
    async def test_request_response_cancelled(
        self, _send_request_mock: AsyncMock
    ) -> None:
        """Test RequestResponse: task cancelled. No callback shall be left."""
        udp_transport = UDPTransport(("192.168.1.1", 0), ("192.168.1.2", 1234))
        requ_resp = RequestResponse(udp_transport, KNXIPBody)
        requ_resp.response_received_event.wait = MagicMock(
            side_effect=asyncio.CancelledError()
        )
        with pytest.raises(asyncio.CancelledError):
            await requ_resp.start()
        # Callback was removed again
        assert not udp_transport.callbacks



"""Example for Date and Time devices."""

import asyncio

from xknx import XKNX
from xknx.devices import TimeDevice


async def main() -> None:
    """Connect to KNX/IP device and broadcast time."""
    xknx = XKNX(daemon_mode=True)
    xknx.devices.async_add(TimeDevice(xknx, "TimeTest", group_address="1/2/3"))
    print("Sending time to KNX bus every hour")
    await xknx.start()
    await xknx.stop()


asyncio.run(main())



"""Unit test for DPT 3 objects."""

from typing import Any

import pytest

from xknx.dpt import DPTArray, DPTBinary, DPTControlBlinds, DPTControlDimming
from xknx.dpt.dpt_1 import Step, UpDown
from xknx.dpt.dpt_3 import ControlBlinds, ControlDimming
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestControlData:
    """Test class for Control data objects."""

    @pytest.mark.parametrize(
        ("data", "dict_value"),
        [
            (
                ControlBlinds(control=UpDown.UP, step_code=0),
                {"control": "up", "step_code": 0},
            ),
            (
                ControlBlinds(control=UpDown.DOWN, step_code=7),
                {"control": "down", "step_code": 7},
            ),
            (
                ControlDimming(control=Step.DECREASE, step_code=5),
                {"control": "decrease", "step_code": 5},
            ),
            (
                ControlDimming(control=Step.INCREASE, step_code=0),
                {"control": "increase", "step_code": 0},
            ),
        ],
    )
    def test_dict(self, data: ControlBlinds, dict_value: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        test_dataclass = data.__class__
        assert test_dataclass.from_dict(dict_value) == data
        assert data.as_dict() == dict_value

    @pytest.mark.parametrize(
        "data",
        [
            {"control": 1, "step_code": "invalid"},
            {"control": "up", "step_code": "invalid"},
            {"control": "increase", "step_code": "invalid"},
            {"control": "down", "step_code": None},
            {"control": None, "step_code": 1},
            {"control": "invalid", "step_code": 0},
            {"control": 2, "step_code": 4},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict with invalid data."""
        with pytest.raises(ValueError):
            ControlBlinds.from_dict(data)
        with pytest.raises(ValueError):
            ControlDimming.from_dict(data)


@pytest.mark.parametrize(
    ("dpt", "dpt_data", "control_enum"),
    [
        (DPTControlDimming, ControlDimming, Step),
        (DPTControlBlinds, ControlBlinds, UpDown),
    ],
)
class TestDPTControlStepCode:
    """Test class for DPT 3 objects."""

    def test_to_knx(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test serializing values."""
        for rawref in range(16):
            control = 1 if rawref >> 3 else 0
            raw = dpt.to_knx(
                dpt_data(
                    control=control_enum(control),
                    step_code=rawref & 0x07,
                )
            )
            assert raw == DPTBinary(rawref)

    def test_to_knx_dict(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test serializing values from dict."""
        for rawref in range(16):
            control = 1 if rawref >> 3 else 0
            raw = dpt.to_knx(
                {
                    "control": control_enum(control).name.lower(),
                    "step_code": rawref & 0x07,
                }
            )
            assert raw == DPTBinary(rawref)

    def test_to_knx_wrong_type(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test serializing wrong type."""
        with pytest.raises(ConversionError):
            dpt.to_knx("")
        with pytest.raises(ConversionError):
            dpt.to_knx(0)

    def test_to_knx_missing_keys(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test serializing map with missing keys."""
        with pytest.raises(ConversionError):
            dpt.to_knx({"control": 0})
        with pytest.raises(ConversionError):
            dpt.to_knx({"step_code": 0})

    def test_to_knx_wrong_value_types(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test serializing map with keys of invalid type."""
        with pytest.raises(ConversionError):
            dpt.to_knx({"control": "", "step_code": 0})
        with pytest.raises(ConversionError):
            dpt.to_knx({"control": 0, "step_code": ""})

    def test_to_knx_wrong_values(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test serializing with invalid values."""
        with pytest.raises(ConversionError):
            dpt.to_knx({"control": -1, "step_code": 0})
        with pytest.raises(ConversionError):
            dpt.to_knx({"control": 2, "step_code": 0})
        with pytest.raises(ConversionError):
            dpt.to_knx({"control": 0, "step_code": -1})
        with pytest.raises(ConversionError):
            dpt.to_knx(dpt_data(control=control_enum(0), step_code=8))

    def test_from_knx(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test parsing from KNX."""
        for raw in range(16):
            control = 1 if raw >> 3 else 0
            valueref = dpt_data(
                control=control_enum(control),
                step_code=raw & 0x07,
            )
            value = dpt.from_knx(DPTBinary((raw,)))
            assert value == valueref

    def test_from_knx_wrong_value(
        self,
        dpt: type[DPTControlDimming | DPTControlBlinds],
        dpt_data: type[ControlDimming | ControlBlinds],
        control_enum: type[Step | UpDown],
    ) -> None:
        """Test parsing invalid values from KNX."""
        with pytest.raises(CouldNotParseTelegram):
            dpt.from_knx(DPTBinary((0x1F,)))
        with pytest.raises(CouldNotParseTelegram):
            dpt.from_knx(DPTArray((1,)))



"""
Module for managing an DPT Switch remote value.

DPT 1.001.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import ConversionError, CouldNotParseTelegram

from .remote_value import GroupAddressesType, RemoteValue, RVCallbackType

if TYPE_CHECKING:
    from xknx.xknx import XKNX


class RemoteValueSwitch(RemoteValue[bool]):
    """Abstraction for remote value of KNX DPT 1.001 / DPT_Switch."""

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "State",
        after_update_cb: RVCallbackType[bool] | None = None,
        invert: bool = False,
    ) -> None:
        """Initialize remote value of KNX DPT 1.001."""
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )
        self.invert = bool(invert)

    def to_knx(self, value: bool) -> DPTBinary:
        """Convert value to payload."""
        if isinstance(value, bool):
            return DPTBinary(value ^ self.invert)
        raise ConversionError(
            "value invalid",
            value=value,
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    def from_knx(self, payload: DPTArray | DPTBinary) -> bool:
        """Convert current payload to value."""
        if payload.value == 0:
            return self.invert
        if payload.value == 1:
            return not self.invert
        raise CouldNotParseTelegram(
            "Payload invalid",
            payload=str(payload),
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    def off(self) -> None:
        """Set value to OFF."""
        self.set(False)

    def on(self) -> None:
        """Set value to ON."""
        # pylint: disable=invalid-name
        self.set(True)



"""Unit test for XKNX Module."""

import logging
from pathlib import Path
import tempfile
from unittest.mock import AsyncMock, Mock, patch

import pytest

from xknx import XKNX
from xknx.exceptions import CommunicationError
from xknx.io import ConnectionConfig, ConnectionType


class TestXknxModule:
    """Test class for XKNX."""

    def test_log_to_file(self) -> None:
        """Test logging enable."""
        with tempfile.TemporaryDirectory() as tmpdir:
            XKNX(log_directory=tmpdir)
            _path = Path(f"{tmpdir}/xknx.log")
            assert _path.is_file()

            # Needed for Windows to release the logging file
            logging.shutdown()
            _path.unlink()

    def test_log_to_file_when_dir_does_not_exist(self) -> None:
        """Test logging enable with non existent directory."""
        XKNX(log_directory="/xknx/is/fun")
        assert not Path("/xknx/is/fun/xknx.log").is_file()

    def test_register_telegram_cb(self) -> None:
        """Test register telegram callback."""
        xknx = XKNX(telegram_received_cb=AsyncMock())
        assert len(xknx.telegram_queue.telegram_received_cbs) == 1

    def test_register_device_cb(self) -> None:
        """Test register telegram callback."""
        xknx = XKNX(device_updated_cb=AsyncMock())
        assert len(xknx.devices.device_updated_cbs) == 1

    def test_register_connection_state_change_cb(self) -> None:
        """Test register con state callback."""
        xknx = XKNX(connection_state_changed_cb=Mock())
        assert len(xknx.connection_manager._connection_state_changed_cbs) == 1

    @patch("xknx.io.KNXIPInterface._start", new_callable=AsyncMock)
    async def test_xknx_start(self, start_mock: AsyncMock) -> None:
        """Test xknx start."""
        xknx = XKNX(state_updater=True)

        await xknx.start()
        start_mock.assert_called_once()
        await xknx.stop()

    @patch("xknx.io.KNXIPInterface._start", new_callable=AsyncMock)
    async def test_xknx_start_as_context_manager(
        self, ipinterface_mock: AsyncMock
    ) -> None:
        """Test xknx start."""
        async with XKNX(state_updater=True) as xknx:
            assert xknx.started.is_set()
            ipinterface_mock.assert_called_once()

    @patch("xknx.io.KNXIPInterface._start", new_callable=AsyncMock)
    async def test_xknx_start_and_stop_with_dedicated_connection_config(
        self, start_mock: AsyncMock
    ) -> None:
        """Test xknx start and stop with connection config."""
        connection_config = ConnectionConfig(connection_type=ConnectionType.TUNNELING)
        xknx = XKNX(connection_config=connection_config)

        await xknx.start()

        start_mock.assert_called_once()
        assert xknx.knxip_interface.connection_config == connection_config

        await xknx.stop()
        assert xknx.knxip_interface._interface is None
        assert xknx.telegram_queue._consumer_task.done()
        assert not xknx.state_updater.started

    @pytest.mark.parametrize(
        "connection_config",
        [
            ConnectionConfig(
                connection_type=ConnectionType.ROUTING, local_ip="127.0.0.1"
            ),
            ConnectionConfig(
                connection_type=ConnectionType.TUNNELING, gateway_ip="127.0.0.2"
            ),
        ],
    )
    @patch(
        "xknx.io.transport.UDPTransport.connect",
        new_callable=AsyncMock,
        side_effect=OSError,
    )
    async def test_xknx_start_initial_connection_error(
        self, transport_connect_mock: AsyncMock, connection_config: ConnectionConfig
    ) -> None:
        """Test xknx start raising when socket can't be set up."""
        xknx = XKNX(
            state_updater=True,
            connection_config=connection_config,
        )
        with pytest.raises(CommunicationError):
            await xknx.start()
        transport_connect_mock.assert_called_once()
        assert xknx.telegram_queue._consumer_task is None  # not started
        assert not xknx.state_updater.started
        assert not xknx.started.is_set()



"""Example on how to read a value from KNX bus."""

import asyncio

from xknx import XKNX
from xknx.core import ValueReader
from xknx.telegram import GroupAddress
from xknx.tools import read_group_value


async def main() -> None:
    """Connect and read value from KNX bus."""
    async with XKNX() as xknx:
        # get the value only (can be decoded when passing `value_type`)
        result = await read_group_value(xknx, "5/1/20")
        print(f"Value: {result}")

        # get the whole telegram
        telegram = await ValueReader(xknx, GroupAddress("5/1/20")).read()
        print(f"Telegram: {telegram}")


asyncio.run(main())



"""Unit test for KNX 4 byte objects."""

import struct
from unittest.mock import patch

import pytest

from xknx.dpt import DPT4ByteSigned, DPT4ByteUnsigned, DPTArray
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPT4Byte:
    """Test class for KNX 4 byte objects."""

    # ####################################################################
    # DPT4ByteUnsigned
    #
    def test_unsigned_settings(self) -> None:
        """Test members of DPT4ByteUnsigned."""
        assert DPT4ByteUnsigned.value_min == 0
        assert DPT4ByteUnsigned.value_max == 4294967295

    def test_unsigned_assert_min_exceeded(self) -> None:
        """Test initialization of DPT4ByteUnsigned with wrong value (Underflow)."""
        with pytest.raises(ConversionError):
            DPT4ByteUnsigned.to_knx(-1)

    def test_unsigned_to_knx_exceed_limits(self) -> None:
        """Test initialization of DPT4ByteUnsigned with wrong value (Overflow)."""
        with pytest.raises(ConversionError):
            DPT4ByteUnsigned.to_knx(4294967296)

    def test_unsigned_value_max_value(self) -> None:
        """Test DPT4ByteUnsigned parsing and streaming."""
        assert DPT4ByteUnsigned.to_knx(4294967295) == DPTArray((0xFF, 0xFF, 0xFF, 0xFF))
        assert (
            DPT4ByteUnsigned.from_knx(DPTArray((0xFF, 0xFF, 0xFF, 0xFF))) == 4294967295
        )

    def test_unsigned_value_min_value(self) -> None:
        """Test parsing and streaming with null values."""
        assert DPT4ByteUnsigned.to_knx(0) == DPTArray((0x00, 0x00, 0x00, 0x00))
        assert DPT4ByteUnsigned.from_knx(DPTArray((0x00, 0x00, 0x00, 0x00))) == 0

    def test_unsigned_value_01234567(self) -> None:
        """Test DPT4ByteUnsigned parsing and streaming."""
        assert DPT4ByteUnsigned.to_knx(19088743) == DPTArray((0x01, 0x23, 0x45, 0x67))
        assert DPT4ByteUnsigned.from_knx(DPTArray((0x01, 0x23, 0x45, 0x67))) == 19088743

    def test_unsigned_wrong_value_from_knx(self) -> None:
        """Test DPT4ByteUnsigned parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPT4ByteUnsigned.from_knx(DPTArray((0xFF, 0x4E, 0x12)))

    # ####################################################################
    # DPT4ByteSigned
    #
    def test_signed_settings(self) -> None:
        """Test members of DPT4ByteSigned."""
        assert DPT4ByteSigned.value_min == -2147483648
        assert DPT4ByteSigned.value_max == 2147483647

    def test_signed_assert_min_exceeded(self) -> None:
        """Test initialization of DPT4ByteSigned with wrong value (Underflow)."""
        with pytest.raises(ConversionError):
            DPT4ByteSigned.to_knx(-2147483649)

    def test_signed_to_knx_exceed_limits(self) -> None:
        """Test initialization of DPT4ByteSigned with wrong value (Overflow)."""
        with pytest.raises(ConversionError):
            DPT4ByteSigned.to_knx(2147483648)

    def test_signed_value_max_value(self) -> None:
        """Test DPT4ByteSigned parsing and streaming."""
        assert DPT4ByteSigned.to_knx(2147483647) == DPTArray((0x7F, 0xFF, 0xFF, 0xFF))
        assert DPT4ByteSigned.from_knx(DPTArray((0x7F, 0xFF, 0xFF, 0xFF))) == 2147483647

    def test_signed_value_min_value(self) -> None:
        """Test DPT4ByteSigned parsing and streaming with null values."""
        assert DPT4ByteSigned.to_knx(-2147483648) == DPTArray((0x80, 0x00, 0x00, 0x00))
        assert (
            DPT4ByteSigned.from_knx(DPTArray((0x80, 0x00, 0x00, 0x00))) == -2147483648
        )

    def test_signed_value_01234567(self) -> None:
        """Test DPT4ByteSigned parsing and streaming."""
        assert DPT4ByteSigned.to_knx(19088743) == DPTArray((0x01, 0x23, 0x45, 0x67))
        assert DPT4ByteSigned.from_knx(DPTArray((0x01, 0x23, 0x45, 0x67))) == 19088743

    def test_signed_wrong_value_from_knx(self) -> None:
        """Test DPT4ByteSigned parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPT4ByteSigned.from_knx(DPTArray((0xFF, 0x4E, 0x12)))

    def test_from_knx_unpack_error(self) -> None:
        """Test DPT4ByteSigned parsing with unpack error."""
        with patch("struct.unpack") as unpack_mock:
            unpack_mock.side_effect = struct.error()
            with pytest.raises(ConversionError):
                DPT4ByteSigned.from_knx(DPTArray((0x01, 0x23, 0x45, 0x67)))

    def test_to_knx_pack_error(self) -> None:
        """Test serializing DPT4ByteSigned with pack error."""
        with patch("struct.pack") as pack_mock:
            pack_mock.side_effect = struct.error()
            with pytest.raises(ConversionError):
                DPT4ByteSigned.to_knx(19088743)



"""Package containing all methods for serialization and deserialization of KNX/IP packets."""

from .body import KNXIPBody, KNXIPBodyResponse
from .connect_request import ConnectRequest, ConnectRequestInformation
from .connect_response import ConnectResponse, ConnectResponseData
from .connectionstate_request import ConnectionStateRequest
from .connectionstate_response import ConnectionStateResponse
from .description_request import DescriptionRequest
from .description_response import DescriptionResponse
from .device_configuration_ack import DeviceConfigurationAck
from .device_configuration_request import DeviceConfigurationRequest
from .dib import (
    DIB,
    DIBDeviceInformation,
    DIBGeneric,
    DIBSecuredServiceFamilies,
    DIBSuppSVCFamilies,
    DIBTunnelingInfo,
)
from .disconnect_request import DisconnectRequest
from .disconnect_response import DisconnectResponse
from .error_code import ErrorCode
from .header import KNXIPHeader
from .hpai import HPAI
from .knxip import KNXIPFrame
from .knxip_enum import (
    ConnectRequestType,
    DIBServiceFamily,
    DIBTypeCode,
    HostProtocol,
    KNXIPServiceType,
    KNXMedium,
    SearchRequestParameterType,
    TunnellingFeatureType,
    TunnellingLayer,
)
from .routing_busy import RoutingBusy
from .routing_indication import RoutingIndication
from .routing_lost_message import RoutingLostMessage
from .search_request import SearchRequest
from .search_request_extended import SearchRequestExtended
from .search_response import SearchResponse
from .search_response_extended import SearchResponseExtended
from .secure_wrapper import SecureWrapper
from .session_authenticate import SessionAuthenticate
from .session_request import SessionRequest
from .session_response import SessionResponse
from .session_status import SessionStatus
from .srp import SRP
from .timer_notify import TimerNotify
from .tunnelling_ack import TunnellingAck
from .tunnelling_feature import (
    TunnellingFeatureGet,
    TunnellingFeatureInfo,
    TunnellingFeatureResponse,
    TunnellingFeatureSet,
)
from .tunnelling_request import TunnellingRequest

__all__ = [
    "DIB",
    "HPAI",
    "SRP",
    "ConnectRequest",
    "ConnectRequestInformation",
    "ConnectRequestType",
    "ConnectResponse",
    "ConnectResponseData",
    "ConnectionStateRequest",
    "ConnectionStateResponse",
    "DIBDeviceInformation",
    "DIBGeneric",
    "DIBSecuredServiceFamilies",
    "DIBServiceFamily",
    "DIBSuppSVCFamilies",
    "DIBTunnelingInfo",
    "DIBTypeCode",
    "DescriptionRequest",
    "DescriptionResponse",
    "DeviceConfigurationAck",
    "DeviceConfigurationRequest",
    "DisconnectRequest",
    "DisconnectResponse",
    "ErrorCode",
    "HostProtocol",
    "KNXIPBody",
    "KNXIPBodyResponse",
    "KNXIPFrame",
    "KNXIPHeader",
    "KNXIPServiceType",
    "KNXMedium",
    "RoutingBusy",
    "RoutingIndication",
    "RoutingLostMessage",
    "SearchRequest",
    "SearchRequestExtended",
    "SearchRequestParameterType",
    "SearchResponse",
    "SearchResponseExtended",
    "SecureWrapper",
    "SessionAuthenticate",
    "SessionRequest",
    "SessionResponse",
    "SessionStatus",
    "TimerNotify",
    "TunnellingAck",
    "TunnellingFeatureGet",
    "TunnellingFeatureInfo",
    "TunnellingFeatureResponse",
    "TunnellingFeatureSet",
    "TunnellingFeatureType",
    "TunnellingLayer",
    "TunnellingRequest",
]



"""Unit test for KNX/IP TunnellingRequest objects."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import (
    ErrorCode,
    KNXIPFrame,
    TunnellingFeatureGet,
    TunnellingFeatureInfo,
    TunnellingFeatureResponse,
    TunnellingFeatureSet,
    TunnellingFeatureType,
)
from xknx.management.application_layer_enum import ReturnCode


class TestKNXIPTunnellingFeature:
    """Test class for KNX/IP TunnelingFeature objects."""

    def test_get(self) -> None:
        """Test parsing and streaming connection tunneling feature get KNX/IP packet."""
        raw = bytes.fromhex("06 10 04 22 00 0c 04 01 17 00 03 00")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, TunnellingFeatureGet)
        assert knxipframe.body.communication_channel_id == 1
        assert knxipframe.body.sequence_counter == 23
        assert (
            knxipframe.body.feature_type == TunnellingFeatureType.BUS_CONNECTION_STATUS
        )
        assert len(knxipframe.body.data) == 0

        tunnelling_request = TunnellingFeatureGet(
            communication_channel_id=1,
            sequence_counter=23,
            feature_type=TunnellingFeatureType.BUS_CONNECTION_STATUS,
        )
        knxipframe2 = KNXIPFrame.init_from_body(tunnelling_request)

        assert knxipframe2.to_knx() == raw

    def test_info(self) -> None:
        """Test parsing and streaming connection tunneling feature info KNX/IP packet."""
        raw = bytes.fromhex("06 10 04 25 00 0e 04 01 17 00 03 00 01 00")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, TunnellingFeatureInfo)
        assert knxipframe.body.communication_channel_id == 1
        assert knxipframe.body.sequence_counter == 23
        assert (
            knxipframe.body.feature_type == TunnellingFeatureType.BUS_CONNECTION_STATUS
        )
        assert knxipframe.body.data == b"\x01\x00"

        tunnelling_request = TunnellingFeatureInfo(
            communication_channel_id=1,
            sequence_counter=23,
            feature_type=TunnellingFeatureType.BUS_CONNECTION_STATUS,
            data=b"\x01\x00",
        )
        knxipframe2 = KNXIPFrame.init_from_body(tunnelling_request)

        assert knxipframe2.to_knx() == raw

    def test_response(self) -> None:
        """Test parsing and streaming connection tunneling feature response KNX/IP packet."""
        raw = bytes.fromhex("06 10 04 23 00 0e 04 01 17 00 03 00 01 00")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, TunnellingFeatureResponse)
        assert knxipframe.body.communication_channel_id == 1
        assert knxipframe.body.sequence_counter == 23
        assert knxipframe.body.status_code == ErrorCode.E_NO_ERROR
        assert (
            knxipframe.body.feature_type == TunnellingFeatureType.BUS_CONNECTION_STATUS
        )
        assert knxipframe.body.return_code == ReturnCode.E_SUCCESS
        assert knxipframe.body.data == b"\x01\x00"

        tunnelling_request = TunnellingFeatureResponse(
            communication_channel_id=1,
            sequence_counter=23,
            feature_type=TunnellingFeatureType.BUS_CONNECTION_STATUS,
            return_code=ReturnCode.E_SUCCESS,
            data=b"\x01\x00",
        )
        knxipframe2 = KNXIPFrame.init_from_body(tunnelling_request)

        assert knxipframe2.to_knx() == raw

    def test_set(self) -> None:
        """Test parsing and streaming connection tunneling feature set KNX/IP packet."""
        raw = bytes.fromhex("06 10 04 24 00 0e 04 01 17 00 08 00 01 00")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, TunnellingFeatureSet)
        assert knxipframe.body.communication_channel_id == 1
        assert knxipframe.body.sequence_counter == 23
        assert (
            knxipframe.body.feature_type
            == TunnellingFeatureType.INTERFACE_FEATURE_INFO_ENABLE
        )
        assert knxipframe.body.data == b"\x01\x00"

        tunnelling_request = TunnellingFeatureSet(
            communication_channel_id=1,
            sequence_counter=23,
            feature_type=TunnellingFeatureType.INTERFACE_FEATURE_INFO_ENABLE,
            data=b"\x01\x00",
        )
        knxipframe2 = KNXIPFrame.init_from_body(tunnelling_request)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_header(self) -> None:
        """Test parsing and streaming wrong Tunnelling Feature (wrong header length byte)."""
        raw = bytes.fromhex("06 10 04 22 00 0c 06 01 17 00 03 00")
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_get_with_data(self) -> None:
        """Test parsing and streaming wrong Get (unexpected data)."""
        raw = bytes.fromhex("06 10 04 22 00 0e 04 01 17 00 03 00 01 00")
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_set_without_data(self) -> None:
        """Test parsing and streaming wrong Get (missing data)."""
        raw = bytes.fromhex("06 10 04 24 00 0c 04 01 17 00 03 00")
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""Module for XKXN Exceptions."""

from __future__ import annotations

import logging
from typing import Any


class XKNXException(Exception):
    """Default XKNX Exception."""

    def __eq__(self, other: object | None) -> bool:
        """Equal operator."""
        return repr(self) == repr(other)

    def __hash__(self) -> int:
        """Hash function."""
        return hash(str(self))

    def __repr__(self) -> str:
        """Representation of object."""
        return str(self)


class CommunicationError(XKNXException):
    """Unable to communicate with KNX bus."""

    def __init__(self, message: str, should_log: bool = True) -> None:
        """Instantiate exception."""
        super().__init__(message)

        self.should_log = should_log


class ConfirmationError(CommunicationError):
    """No confirmation received from KNX server for sent Telegram."""


class TunnellingAckError(CommunicationError):
    """No ACK or error status received from UDP KNX server for sent Telegram."""


class IPSecureError(CommunicationError):
    """Error in IP Secure communication."""


class CouldNotParseTelegram(XKNXException):
    """Could not parse telegram error."""

    def __init__(self, description: str, **kwargs: Any) -> None:
        """Initialize CouldNotParseTelegram class."""
        super().__init__()
        self.description = description
        self.parameter = kwargs

    def _format_parameter(self) -> str:
        return " ".join(
            [f"{key}={value!r}" for (key, value) in sorted(self.parameter.items())]
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            "<CouldNotParseTelegram "
            f'description="{self.description}" {self._format_parameter()}/>'
        )


class CouldNotParseKNXIP(XKNXException):
    """Exception class for wrong KNXIP data."""

    def __init__(self, description: str = "") -> None:
        """Initialize CouldNotParseKNXIP class."""
        super().__init__()
        self.description = description

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<CouldNotParseKNXIP description="{self.description}" />'


class KNXSecureValidationError(CouldNotParseKNXIP):
    """Exception class for invalid KNX Secure data."""

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<KNXSecureValidationError description="{self.description}" />'


class IncompleteKNXIPFrame(CouldNotParseKNXIP):
    """
    Exception class for incomplete KNXIP data.

    Used for TCP connections to indicate to buffer the data until the complete frame is received.
    UDP connections should just handle CouldNotParseKNXIP.
    """

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<IncompleteKNXIPFrame description="{self.description}" />'


class CouldNotParseCEMI(XKNXException):
    """Exception class for wrong CEMI data."""

    def __init__(self, description: str = "") -> None:
        """Initialize CouldNotParseCEMI class."""
        super().__init__()
        self.description = description

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<CouldNotParseCEMI description="{self.description}" />'


class UnsupportedCEMIMessage(XKNXException):
    """Exception class for unsupported CEMI Messages."""

    def __init__(self, description: str = "") -> None:
        """Initialize UnsupportedCEMIMessage class."""
        super().__init__()
        self.description = description

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<UnsupportedCEMIMessage description="{self.description}" />'


class ConversionError(XKNXException):
    """Exception class for error while converting one type to another."""

    def __init__(self, description: str, **kwargs: Any) -> None:
        """Initialize ConversionError class."""
        super().__init__()
        self.description = description
        self.parameter = kwargs

    def _format_parameter(self) -> str:
        return " ".join(
            [f"{key}={value!r}" for (key, value) in sorted(self.parameter.items())]
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<ConversionError description="{self.description}" {self._format_parameter()}/>'


class CouldNotParseAddress(XKNXException):
    """Exception class for wrong address format."""

    def __init__(self, address: Any = None, message: str = "") -> None:
        """Initialize CouldNotParseAddress class."""
        super().__init__()
        self.address = address
        self.message = message

    def __str__(self) -> str:
        """Return object as readable string."""
        _msg = f'message="{self.message}" ' if self.message else ""
        return f"<CouldNotParseAddress address={self.address!r} {_msg}/>"


class DeviceIllegalValue(XKNXException):
    """Exception class for setting a value of a device with an illegal value."""

    def __init__(self, description: str, value: Any) -> None:
        """Initialize DeviceIllegalValue class."""
        super().__init__()
        self.value = value
        self.description = description

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<DeviceIllegalValue description="{self.description}" value={self.value!r} />'


class DataSecureError(XKNXException):
    """Exception class for KNX Data Secure handling."""

    def __init__(self, message: str, log_level: int = logging.WARNING) -> None:
        """Instantiate exception."""
        super().__init__(message)
        self.log_level = log_level


class InvalidSecureConfiguration(XKNXException):
    """Exception class used when the secure configuration is invalid."""


class ManagementConnectionError(XKNXException):
    """Exception class used when a management connection fails."""


class ManagementConnectionRefused(ManagementConnectionError):
    """Exception class used when a management connection request is refused."""


class ManagementConnectionTimeout(ManagementConnectionError):
    """Exception class used when a management connection timed out."""
